import asyncio
import json
import os
from collections import namedtuple
from datetime import timedelta

import requests
from nostr_sdk import Tag, Keys, nip44_encrypt, nip44_decrypt, Nip44Version, EventBuilder, Client, Filter, Kind, \
    EventId, nip04_decrypt, nip04_encrypt, Options, NostrSigner, PublicKey, Metadata

from nostr_dvm.utils.database_utils import fetch_user_metadata
from nostr_dvm.utils.definitions import EventDefinitions, relay_timeout, relay_timeout_long
from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nostr_utils import check_and_set_private_key
from nostr_dvm.utils.print_utils import bcolors
from nostr_dvm.utils.zap_utils import pay_bolt11_ln_bits, zaprequest


class NutWallet(object):
    def __init__(self):
        self.name: str = "NutWallet"
        self.description: str = ""
        self.balance: int = 0
        self.unit: str = "sat"
        self.mints: list = []
        self.relays: list = []
        self.nutmints: list = []
        self.privkey: str = ""
        self.d: str = ""
        self.a: str = ""
        self.legacy_encryption: bool = False  # Use Nip04 instead of Nip44, for reasons, turn to False ASAP.
        self.trust_unknown_mints: bool = False
        self.missing_balance_strategy: str = "mint"  # none to do nothing until manually minted, mint to mint from lightning or swap to use existing tokens from other mints (fees!) (not working yet!)


class NutMint(object):
    def __init__(self):
        self.previous_event_id = None
        self.proofs: list = []
        self.mint_url: str = ""
        self.previous_event_id: EventId
        self.a: str = ""

    def available_balance(self):
        balance = 0
        for proof in self.proofs:
            balance += proof.amount
        return balance


class NutZapWallet:

    async def client_connect(self, relay_list):
        keys = Keys.parse(check_and_set_private_key("TEST_ACCOUNT_PK"))
        wait_for_send = False
        skip_disconnected_relays = True
        opts = (Options().wait_for_send(wait_for_send).send_timeout(timedelta(seconds=5))
                .skip_disconnected_relays(skip_disconnected_relays))

        signer = NostrSigner.keys(keys)
        client = Client.with_opts(signer, opts)
        for relay in relay_list:
            await client.add_relay(relay)
        await client.connect()
        return client, keys

    async def create_new_nut_wallet(self, mint_urls, relays, client, keys, name, description):
        new_nut_wallet = NutWallet()
        new_nut_wallet.privkey = Keys.generate().secret_key().to_hex()
        new_nut_wallet.balance = 0
        new_nut_wallet.unit = "sats"
        new_nut_wallet.name = name
        new_nut_wallet.description = description
        new_nut_wallet.mints = mint_urls
        new_nut_wallet.relays = relays
        new_nut_wallet.d = "wallet"
        new_nut_wallet.a = str(Kind(7375).as_u16()) + ":" + keys.public_key().to_hex() + ":" + new_nut_wallet.d
        print("Creating Wallet..")
        send_response_id = await self.create_or_update_nut_wallet_event(new_nut_wallet, client, keys)

        if send_response_id is None:
            print("Warning: Not published")

        print(new_nut_wallet.name + ": " + str(new_nut_wallet.balance) + " " + new_nut_wallet.unit + " Mints: " + str(
            new_nut_wallet.mints) + " Key: " + new_nut_wallet.privkey)

    async def create_or_update_nut_wallet_event(self, nut_wallet: NutWallet, client, keys):
        innertags = [Tag.parse(["balance", str(nut_wallet.balance), nut_wallet.unit]).as_vec(),
                     Tag.parse(["privkey", nut_wallet.privkey]).as_vec()]

        if nut_wallet.legacy_encryption:
            content = nip04_encrypt(keys.secret_key(), keys.public_key(), json.dumps(innertags))
        else:
            content = nip44_encrypt(keys.secret_key(), keys.public_key(), json.dumps(innertags), Nip44Version.V2)

        if nut_wallet.unit is None:
            nut_wallet.unit = "sat"

        tags = [Tag.parse(["name", nut_wallet.name]),
                Tag.parse(["unit", nut_wallet.unit]),
                Tag.parse(["description", nut_wallet.description]),
                Tag.parse(["d", nut_wallet.d])]

        for mint in nut_wallet.mints:
            mint_tag = Tag.parse(["mint", mint])
            tags.append(mint_tag)

        for relay in nut_wallet.relays:
            relay_tag = Tag.parse(["relay", relay])
            tags.append(relay_tag)

        event = EventBuilder(EventDefinitions.KIND_NUT_WALLET, content, tags).to_event(keys)
        send_response = await client.send_event(event)

        print(
            bcolors.BLUE + "[" + nut_wallet.name + "] announced nut wallet (" + send_response.id.to_hex() + ")" + bcolors.ENDC)
        return send_response.id

    async def get_nut_wallet(self, client, keys) -> NutWallet:
        from cashu.core.base import Proof
        nut_wallet = None

        wallet_filter = Filter().kind(EventDefinitions.KIND_NUT_WALLET).author(keys.public_key())
        # relay_timeout = EventSource.relays(timedelta(seconds=10))
        wallets = await client.get_events_of([wallet_filter], relay_timeout_long)

        if len(wallets) > 0:

            nut_wallet = NutWallet()

            latest = 0
            best_wallet = None
            for wallet_event in wallets:

                isdeleted = False
                for tag in wallet_event.tags():
                    if tag.as_vec()[0] == "deleted":
                        isdeleted = True
                        break
                if isdeleted:
                    continue
                else:
                    if wallet_event.created_at().as_secs() > latest:
                        latest = wallet_event.created_at().as_secs()
                        best_wallet = wallet_event

            try:
                content = nip44_decrypt(keys.secret_key(), keys.public_key(), best_wallet.content())
                print(content)
            except:
                content = nip04_decrypt(keys.secret_key(), keys.public_key(), best_wallet.content())
                print(content)
                print("Warning: This Wallet is using a NIP04 enconding.., it should use NIP44 encoding ")
                nut_wallet.legacy_encryption = True

            inner_tags = json.loads(content)
            for tag in inner_tags:
                # These tags must be encrypted instead of in the outer tags
                if tag[0] == "balance":
                    nut_wallet.balance = int(tag[1])
                elif tag[0] == "privkey":
                    nut_wallet.privkey = tag[1]

                # These tags can be encrypted instead of in the outer tags
                elif tag[0] == "name":
                    nut_wallet.name = tag[1]
                elif tag[0] == "description":
                    nut_wallet.description = tag[1]
                elif tag[0] == "unit":
                    nut_wallet.unit = tag[1]
                elif tag[0] == "relay":
                    if tag[1] not in nut_wallet.relays:
                        nut_wallet.relays.append(tag[1])
                elif tag[0] == "mint":
                    if tag[1] not in nut_wallet.mints:
                        nut_wallet.mints.append(tag[1])

            for tag in best_wallet.tags():
                if tag.as_vec()[0] == "d":
                    nut_wallet.d = tag.as_vec()[1]

                # These tags can be in the outer tags (if not encrypted)
                elif tag.as_vec()[0] == "name":
                    nut_wallet.name = tag.as_vec()[1]
                elif tag.as_vec()[0] == "description":
                    nut_wallet.description = tag.as_vec()[1]
                elif tag.as_vec()[0] == "unit":
                    nut_wallet.unit = tag.as_vec()[1]
                elif tag.as_vec()[0] == "relay":
                    if tag.as_vec()[1] not in nut_wallet.relays:
                        nut_wallet.relays.append(tag.as_vec()[1])
                elif tag.as_vec()[0] == "mint":
                    if tag.as_vec()[1] not in nut_wallet.mints:
                        nut_wallet.mints.append(tag.as_vec()[1])
            nut_wallet.a = str("37375:" + best_wallet.author().to_hex() + ":" + nut_wallet.d)

            # Now all proof events
            proof_filter = Filter().kind(Kind(7375)).author(keys.public_key())
            # relay_timeout = EventSource.relays(timedelta(seconds=5))
            proof_events = await client.get_events_of([proof_filter], relay_timeout)

            latest_proof_sec = 0
            latest_proof_event_id = EventId
            for proof_event in proof_events:
                if proof_event.created_at().as_secs() > latest_proof_sec:
                    latest_proof_sec = proof_event.created_at().as_secs()
                    latest_proof_event_id = proof_event.id()

            for proof_event in proof_events:
                try:
                    content = nip44_decrypt(keys.secret_key(), keys.public_key(), proof_event.content())
                except:
                    content = nip04_decrypt(keys.secret_key(), keys.public_key(), proof_event.content())
                    print("Warning: This Proofs event is using a NIP04 enconding.., it should use NIP44 encoding ")

                proofs_json = json.loads(content)
                mint_url = ""
                a = ""
                print("")
                print("AVAILABLE MINT:")

                try:
                    mint_url = proofs_json['mint']
                    print("mint: " + mint_url)
                    a = proofs_json['a']
                    print("a: " + a)
                except Exception as e:
                    pass

                for tag in proof_event.tags():
                    if tag.as_vec()[0] == "mint":
                        mint_url = tag.as_vec()[1]
                        print("mint: " + mint_url)
                    elif tag.as_vec()[0] == "a":
                        a = tag.as_vec()[1]
                        print("a: " + a)

                nut_mint = NutMint()
                nut_mint.mint_url = mint_url
                nut_mint.a = a
                nut_mint.previous_event_id = latest_proof_event_id
                nut_mint.proofs = []

                for proof in proofs_json['proofs']:
                    proofs = [x for x in nut_mint.proofs if x.secret == proof['secret']]
                    if len(proofs) == 0:
                        nut_proof = Proof()
                        nut_proof.id = proof['id']
                        nut_proof.secret = proof['secret']
                        nut_proof.amount = proof['amount']
                        nut_proof.C = proof['C']
                        nut_mint.proofs.append(nut_proof)
                        # print(proof)

                mints = [x for x in nut_wallet.nutmints if x.mint_url == mint_url]
                if len(mints) == 0:
                    nut_wallet.nutmints.append(nut_mint)
                print("Mint Balance: " + str(nut_mint.available_balance()) + " Sats")

        return nut_wallet

    async def update_nut_wallet(self, nut_wallet, mints, client, keys):
        for mint in mints:
            if mint not in nut_wallet.mints:
                nut_wallet.mints.append(mint)

        balance = 0
        for mint in nut_wallet.nutmints:
            for proof in mint.proofs:
                balance += proof.amount

        nut_wallet.balance = balance

        id = await self.create_or_update_nut_wallet_event(nut_wallet, client, keys)

        if id is None:
            print(bcolors.RED + str("Error publishing WalletEvent") + bcolors.ENDC)

        print(nut_wallet.name + ": " + str(nut_wallet.balance) + " " + nut_wallet.unit + " Mints: " + str(
            nut_wallet.mints) + " Key: " + nut_wallet.privkey)

        return nut_wallet

    def get_mint(self, nut_wallet, mint_url) -> NutMint:
        mints = [x for x in nut_wallet.nutmints if x.mint_url == mint_url]
        if len(mints) == 0:
            mint = NutMint()
            mint.proofs = []
            mint.previous_event_id = None
            mint.a = nut_wallet.a
            mint.mint_url = mint_url

        else:
            mint = mints[0]

        return mint

    async def create_transaction_history_event(self, nut_wallet: NutWallet, amount: int, unit: str,
                                               event_old: EventId | None,
                                               event_new: EventId, direction: str, marker, sender_hex, event_hex,
                                               client: Client, keys: Keys):
        # direction
        # in = received
        # out = sent

        # marker:
        # created - A new token event was created.
        # destroyed - A token event was destroyed.
        # redeemed - A [[NIP-61]] nutzap was redeemed."

        relays = await client.relays()
        relay_hints = relays.keys()
        relay_hint = list(relay_hints)[0]

        inner_tags = []
        inner_tags.append(["direction", direction])
        inner_tags.append(["amount", str(amount), unit])

        if event_old is not None:
            inner_tags.append(["e", event_old.to_hex(), relay_hint, "destroyed"])

        inner_tags.append(["e", event_new.to_hex(), relay_hint, "created"])

        message = json.dumps(inner_tags)
        if nut_wallet.legacy_encryption:
            content = nip04_encrypt(keys.secret_key(), keys.public_key(), message)
        else:
            content = nip44_encrypt(keys.secret_key(), keys.public_key(), message, Nip44Version.V2)

        tags = [Tag.parse(["a", nut_wallet.a])]
        if marker == "redeemed" or marker == "zapped":
            e_tag = Tag.parse(["e", event_hex, relay_hint, marker])
            tags.append(e_tag)
            p_tag = Tag.parse(["p", sender_hex])
            tags.append(p_tag)

        event = EventBuilder(Kind(7376), content, tags).to_event(keys)
        eventid = await client.send_event(event)

    async def create_unspent_proof_event(self, nut_wallet: NutWallet, mint_proofs, mint_url, amount, direction, marker,
                                         sender_hex, event_hex,
                                         client, keys):
        new_proofs = []
        mint = self.get_mint(nut_wallet, mint_url)
        mint.proofs = mint_proofs
        for proof in mint_proofs:
            proofjson = {
                "id": proof['id'],
                "amount": proof['amount'],
                "secret": proof['secret'],
                "C": proof['C']
            }
            # print("Mint proofs:")
            # print(proof)
            new_proofs.append(proofjson)
        old_event_id = mint.previous_event_id

        if mint.previous_event_id is not None:
            print(
                bcolors.MAGENTA + "[" + nut_wallet.name + "] Deleted previous proofs event.. : (" + mint.previous_event_id.to_hex() + ")" + bcolors.ENDC)
            evt = EventBuilder.delete([mint.previous_event_id], reason="deleted").to_event(
                keys)  # .to_pow_event(keys, 28)
            response = await client.send_event(evt)

        tags = []
        # print(nut_wallet.a)
        a_tag = Tag.parse(["a", nut_wallet.a])
        tags.append(a_tag)

        j = {
            "mint": mint_url,
            "proofs": new_proofs
        }

        message = json.dumps(j)

        # print(message)
        if nut_wallet.legacy_encryption:
            content = nip04_encrypt(keys.secret_key(), keys.public_key(), message)
        else:
            content = nip44_encrypt(keys.secret_key(), keys.public_key(), message, Nip44Version.V2)

        event = EventBuilder(Kind(7375), content, tags).to_event(keys)
        eventid = await client.send_event(event)
        await self.create_transaction_history_event(nut_wallet, amount, nut_wallet.unit, old_event_id, eventid.id,
                                                    direction, marker, sender_hex, event_hex, client, keys)

        print(
            bcolors.GREEN + "[" + nut_wallet.name + "] Published new proofs event.. : (" + eventid.id.to_hex() + ")" + bcolors.ENDC)

        return eventid.id

    async def mint_token(self, mint, amount):
        from cashu.wallet.wallet import Wallet
        # TODO probably there's a library function for this
        url = mint + "/v1/mint/quote/bolt11"
        json_object = {"unit": "sat", "amount": amount}

        headers = {"Content-Type": "application/json; charset=utf-8"}
        request_body = json.dumps(json_object).encode('utf-8')
        request = requests.post(url, data=request_body, headers=headers)
        tree = json.loads(request.text)

        lnbits_config = {
            "LNBITS_ADMIN_KEY": os.getenv("LNBITS_ADMIN_KEY"),
            "LNBITS_URL": os.getenv("LNBITS_HOST")
        }
        lnbits_config_obj = namedtuple("LNBITSCONFIG", lnbits_config.keys())(*lnbits_config.values())

        paymenthash = pay_bolt11_ln_bits(tree["request"], lnbits_config_obj)
        print(paymenthash)
        url = f"{mint}/v1/mint/quote/bolt11/{tree['quote']}"

        response = requests.get(url, data=request_body, headers=headers)
        tree2 = json.loads(response.text)
        waitfor = 5
        while not tree2["paid"]:
            await asyncio.sleep(1)
            response = requests.get(url, data=request_body, headers=headers)
            tree2 = json.loads(response.text)
            waitfor -= 1
            if waitfor == 0:
                break

        if tree2["paid"]:
            # print(response.text)
            wallet = await Wallet.with_db(
                url=mint,
                db="db/Cashu",
            )

            await wallet.load_mint()
            proofs = await wallet.mint(amount, tree['quote'], None)
            return proofs

    async def announce_nutzap_info_event(self, nut_wallet, client, keys):
        tags = []
        for relay in nut_wallet.relays:
            tags.append(Tag.parse(["relay", relay]))
        for mint in nut_wallet.mints:
            tags.append(Tag.parse(["mint", mint]))

        pubkey = Keys.parse(nut_wallet.privkey).public_key().to_hex()
        tags.append(Tag.parse(["pubkey", pubkey]))

        event = EventBuilder(Kind(10019), "", tags).to_event(keys)
        eventid = await client.send_event(event)
        print(
            bcolors.CYAN + "[" + nut_wallet.name + "] Announced mint preferences info event (" + eventid.id.to_hex() + ")" + bcolors.ENDC)

    async def fetch_mint_info_event(self, pubkey, client):
        mint_info_filter = Filter().kind(Kind(10019)).author(PublicKey.parse(pubkey))
        # relay_timeout = EventSource.relays(timedelta(seconds=5))
        preferences = await client.get_events_of([mint_info_filter], relay_timeout)
        mints = []
        relays = []
        pubkey = ""

        if len(preferences) > 0:
            preference = preferences[0]

            for tag in preference.tags():
                if tag.as_vec()[0] == "pubkey":
                    pubkey = tag.as_vec()[1]
                elif tag.as_vec()[0] == "relay":
                    relays.append(tag.as_vec()[1])
                elif tag.as_vec()[0] == "mint":
                    mints.append(tag.as_vec()[1])

        return pubkey, mints, relays

    async def update_spend_mint_proof_event(self, nut_wallet, send_proofs, mint_url, marker, sender_hex, event_hex,
                                            client, keys):
        mint = self.get_mint(nut_wallet, mint_url)

        print(mint.mint_url)
        print(send_proofs)
        amount = 0
        for send_proof in send_proofs:
            entry = [x for x in mint.proofs if x.id == send_proof.id and x.secret == send_proof.secret]
            if len(entry) > 0:
                mint.proofs.remove(entry[0])
                amount += send_proof.amount

        # create new event
        mint.previous_event_id = await self.create_unspent_proof_event(nut_wallet, mint.proofs, mint.mint_url, amount,
                                                                       "out",
                                                                       marker, sender_hex,
                                                                       event_hex, client, keys)
        nut_wallet.balance = nut_wallet.balance - amount
        return await self.update_nut_wallet(nut_wallet, [mint.mint_url], client, keys)

    async def mint_cashu(self, nut_wallet: NutWallet, mint_url, client, keys, amount):
        print("Minting new tokens on: " + mint_url)
        # Mint the Token at the selected mint
        proofs = await self.mint_token(mint_url, amount)
        print(proofs)

        return await self.add_proofs_to_wallet(nut_wallet, mint_url, proofs, "created", None, None, client, keys)

    async def add_proofs_to_wallet(self, nut_wallet, mint_url, new_proofs, marker, sender, event, client: Client,
                                   keys: Keys):
        mint = self.get_mint(nut_wallet, mint_url)
        additional_amount = 0
        # store the new proofs in proofs_temp
        all_proofs = []
        # check for other proofs from same mint, add them to the list of proofs
        for nut_proof in mint.proofs:
            all_proofs.append(nut_proof)
        # add new proofs and calculate additional balance
        for proof in new_proofs:
            all_proofs.append(proof)
            nut_wallet.balance += proof.amount
            additional_amount += proof.amount

        print("New amount: " + str(additional_amount))
        mint.previous_event_id = await self.create_unspent_proof_event(nut_wallet, all_proofs, mint_url,
                                                                       additional_amount, "in",
                                                                       marker,
                                                                       sender, event, client, keys)

        return await self.update_nut_wallet(nut_wallet, [mint_url], client, keys)

    async def handle_low_balance_on_mint(self, nut_wallet, mint_to_send, mint, amount, client, keys):

        required_amount = amount - mint.available_balance()

        if nut_wallet.missing_balance_strategy == "mint":
            await self.mint_cashu(nut_wallet, mint_to_send, client, keys, required_amount)

        elif nut_wallet.missing_balance_strategy == "swap":
            for nutmint in nut_wallet.nutmints:
                estimated_fees = 3
                if nutmint.available_balance() > required_amount + estimated_fees:
                    print(nutmint.mint_url)
                    await self.swap(required_amount, mint_to_send, nutmint.mint_url, nut_wallet)
                    break
        else:
            print(bcolors.RED + "[" + nut_wallet.name + "] Not enough Balance on Mint, mint some tokens first. " + str(
                amount) + " " + nut_wallet.unit + bcolors.ENDC)

    async def send_nut_zap(self, amount, comment, nut_wallet: NutWallet, zapped_event, zapped_user, client: Client,
                           keys: Keys):
        from cashu.wallet.wallet import Wallet
        unit = "sats"

        p2pk_pubkey, mints, relays = await self.fetch_mint_info_event(zapped_user, client)
        if len(mints) == 0:
            print("No preferred mint set, returning")
            return

        mint_success = False
        index = 0
        mint_url = ""
        sufficent_budget = False

        # Some logic.
        # First look if we have balance on a mint the user has in their list of trusted mints and use it
        for mint in nut_wallet.nutmints:
            if mint.available_balance() >= amount and mint.mint_url in mints:
                mint_url = mint.mint_url
                sufficent_budget = True
                break
        # If that's not the case, lets look or mints we both trust, take the first one.
        if not sufficent_budget:
            mint_url = next(i for i in nut_wallet.mints if i in mints)
            mint = self.get_mint(nut_wallet, mint_url)
            if mint.available_balance() < amount:
                await self.handle_low_balance_on_mint(nut_wallet, mint_url, mint, amount, client, keys)

            # If that's not the case, iterate over the recipents mints and try to mint there. This might be a bit dangerous as not all mints might give cashu, so loss of ln is possible
            if mint_url is None:
                if nut_wallet.trust_unknown_mints:
                    # maybe we don't do this for now..
                    while not mint_success:
                        try:
                            mint_url = mints[index]  #
                            # Maybe we introduce a list of known failing mints..
                            if mint_url == "https://stablenut.umint.cash":
                                raise Exception("stablemint bad")
                            mint = self.get_mint(nut_wallet, mint_url)
                            if mint.available_balance() < amount:
                                await self.handle_low_balance_on_mint(nut_wallet, mint_url, mint, amount, client, keys)
                            mint_success = True
                        except:
                            mint_success = False
                            index += 1
                else:
                    print("No trusted mints founds, enable trust_unknown_mints if you still want to proceed...")
                    return

        tags = [Tag.parse(["amount", str(amount)]),
                Tag.parse(["unit", unit]),
                Tag.parse(["u", mint_url]),
                Tag.parse(["p", zapped_user])]

        if zapped_event != "" and zapped_event is not None:
            tags.append(Tag.parse(["e", zapped_event]))

        mint = self.get_mint(nut_wallet, mint_url)

        cashu_wallet = await Wallet.with_db(
            url=mint_url,
            db="db/Cashu",
            name="wallet_mint_api",
        )

        await cashu_wallet.load_mint()
        secret_lock = await cashu_wallet.create_p2pk_lock("02" + p2pk_pubkey)  # sender side

        try:
            proofs, fees = await cashu_wallet.select_to_send(mint.proofs, amount)
            print(fees)
            keep_proofs, send_proofs = await cashu_wallet.swap_to_send(
                proofs, amount, secret_lock=secret_lock, set_reserved=True
            )

            print(keep_proofs)

            for proof in send_proofs:
                nut_proof = {
                    'id': proof.id,
                    'C': proof.C,
                    'amount': proof.amount,
                    'secret': proof.secret,
                }
                tags.append(Tag.parse(["proof", json.dumps(nut_proof)]))

            event = EventBuilder(Kind(9321), comment, tags).to_event(keys)
            response = await client.send_event(event)

            await self.update_spend_mint_proof_event(nut_wallet, proofs, mint_url, "zapped", keys.public_key().to_hex(),
                                                     response.id.to_hex(), client, keys)

            print(bcolors.YELLOW + "[" + nut_wallet.name + "] Sent NutZap 🥜️⚡ with " + str(
                amount) + " " + nut_wallet.unit + " to "
                  + PublicKey.parse(zapped_user).to_bech32() +
                  "(" + response.id.to_hex() + ")" + bcolors.ENDC)

        except Exception as e:
            print(e)

    def print_transaction_history(self, transactions, keys):
        for transaction in transactions:
            try:
                content = nip44_decrypt(keys.secret_key(), keys.public_key(), transaction.content())
            except:
                content = nip04_decrypt(keys.secret_key(), keys.public_key(), transaction.content())

            innertags = json.loads(content)
            direction = ""
            amount = ""
            unit = "sats"
            for tag in innertags:
                if tag[0] == "direction":
                    direction = tag[1]
                elif tag[0] == "amount":
                    amount = tag[1]
                    unit = tag[2]
                    if amount == "1" and unit == "sats":
                        unit = "sat"
            sender = ""
            event = ""
            for tag in transaction.tags():
                if tag.as_vec()[0] == "p":
                    sender = tag.as_vec()[1]
                elif tag.as_vec()[0] == "e":
                    event = tag.as_vec()[1]
                    # marker = tag.as_vec()[2]

            if direction == "in":
                color = bcolors.GREEN
                action = "minted"
                dir = "from"
            else:
                color = bcolors.RED
                action = "spent"
                dir = "to"

            if sender != "" and event != "":
                print(
                    color + f"{direction:3}" + " " + f"{amount:6}" + " " + unit + " at " + transaction.created_at().to_human_datetime().replace(
                        "T", " ").replace("Z",
                                          " ") + "GMT" + bcolors.ENDC + " " + bcolors.YELLOW + " (Nutzap 🥜⚡️ " + dir + ": " + PublicKey.parse(
                        sender).to_bech32() + "(" + event + "))" + bcolors.ENDC)
            else:
                print(
                    color + f"{direction:3}" + " " + f"{amount:6}" + " " + unit + " at " + transaction.created_at().to_human_datetime().replace(
                        "T", " ").replace("Z", " ") + "GMT" + " " + " (" + action + ")" + bcolors.ENDC)

    async def reedeem_nutzap(self, event, nut_wallet: NutWallet, client: Client, keys: Keys):
        from cashu.wallet.wallet import Wallet
        from cashu.core.base import Proof
        from cashu.core.crypto.keys import PrivateKey
        try:
            proofs = []
            mint_url = ""
            amount = 0
            unit = "sat"
            zapped_user = ""
            zapped_event = ""
            sender = event.author().to_hex()
            message = event.content()
            for tag in event.tags():
                if tag.as_vec()[0] == "proof":
                    proof_json = json.loads(tag.as_vec()[1])
                    proof = Proof().from_dict(proof_json)
                    proofs.append(proof)
                elif tag.as_vec()[0] == "u":
                    mint_url = tag.as_vec()[1]
                elif tag.as_vec()[0] == "amount":
                    amount = int(tag.as_vec()[1])
                elif tag.as_vec()[0] == "unit":
                    unit = tag.as_vec()[1]
                elif tag.as_vec()[0] == "p":
                    zapped_user = tag.as_vec()[1]
                elif tag.as_vec()[0] == "e":
                    zapped_event = tag.as_vec()[1]

            cashu_wallet = await Wallet.with_db(
                url=mint_url,
                db="db/Receiver",
                name="receiver",
            )
            cashu_wallet.private_key = PrivateKey(bytes.fromhex(nut_wallet.privkey), raw=True)
            await cashu_wallet.load_mint()

            new_proofs, _ = await cashu_wallet.redeem(proofs)
            mint = self.get_mint(nut_wallet, mint_url)
            print(mint.proofs)
            print(new_proofs)
            count_amount = 0
            for proof in new_proofs:
                count_amount += proof.amount
            await self.add_proofs_to_wallet(nut_wallet, mint_url, new_proofs, "redeemed", event.author().to_hex(),
                                            event.id().to_hex(), client, keys)

            return count_amount, message, sender
        except Exception as e:
            print(bcolors.RED + str(e) + bcolors.ENDC)
            return None, message, sender

    async def melt_cashu(self, nut_wallet, mint_url, total_amount, client, keys, lud16=None, npub=None):
        from cashu.wallet.wallet import Wallet
        mint = self.get_mint(nut_wallet, mint_url)

        cashu_wallet = await Wallet.with_db(
            url=mint_url,
            db="db/Cashu",
            name="wallet_mint_api",
        )
        await cashu_wallet.load_mint()
        cashu_wallet.proofs = mint.proofs

        estimated_fees = max(int(total_amount * 0.02), 3)
        estimated_redeem_invoice_amount = total_amount - estimated_fees

        if npub is None:
            # if we don't pass the npub, we default to our pubkey
            npub = Keys.parse(check_and_set_private_key("TEST_ACCOUNT_PK")).public_key().to_hex()

        if lud16 is None:
            # if we don't pass a lud16, we try to fetch one from our profile (make sure it's set)
            name, nip05, lud16 = await fetch_user_metadata(npub, client)

        invoice = zaprequest(lud16, estimated_redeem_invoice_amount, "Melting from your nutsack", None,
                             PublicKey.parse(npub), keys, DVMConfig().RELAY_LIST, zaptype="private")
        # else:
        #    invoice = create_bolt11_lud16(lud16, estimated_redeem_invoice_amount)
        quote = await cashu_wallet.melt_quote(invoice)

        send_proofs, _ = await cashu_wallet.select_to_send(cashu_wallet.proofs, total_amount)
        await cashu_wallet.melt(send_proofs, invoice, estimated_fees, quote.quote)
        await self.update_spend_mint_proof_event(nut_wallet, send_proofs, mint_url, "", None,
                                                 None, client, keys)

        print(bcolors.YELLOW + "[" + nut_wallet.name + "] Redeemed on Lightning ⚡ " + str(
            total_amount - estimated_fees) + " (Fees: " + str(estimated_fees) + ") " + nut_wallet.unit
              + bcolors.ENDC)

    async def swap(self, amountinsats, incoming_mint_url, outgoing_mint_url, nut_wallet):
        # TODO this doesn't seem to work yet.
        from cashu.wallet.cli.cli_helpers import print_mint_balances
        from cashu.wallet.wallet import Wallet
        from cashu.core.crypto.keys import PrivateKey

        outgoing_mint = self.get_mint(nut_wallet, outgoing_mint_url)

        outgoing_wallet = await Wallet.with_db(
            url=outgoing_mint_url,
            db="db/Cashu",
            name="outgoing",
        )
        outgoing_wallet.private_key = PrivateKey(bytes.fromhex(nut_wallet.privkey), raw=True)
        await outgoing_wallet.load_mint()
        outgoing_wallet.proofs = outgoing_mint.proofs

        print(outgoing_wallet.available_balance)

        incoming_mint = self.get_mint(nut_wallet, incoming_mint_url)
        incoming_wallet = await Wallet.with_db(
            url=incoming_mint_url,
            db="db/Cashu",
            name="incoming",
        )
        incoming_wallet.private_key = PrivateKey(bytes.fromhex(nut_wallet.privkey), raw=True)
        await incoming_wallet.load_mint()
        incoming_wallet.proofs = incoming_mint.proofs

        if incoming_wallet.url == outgoing_wallet.url:
            raise Exception("mints for swap have to be different")

        print("Incoming Mint units: " + incoming_wallet.unit.name)
        assert amountinsats > 0, "amount is not positive"

        # request invoice from incoming mint
        invoice = await incoming_wallet.request_mint(amountinsats)

        # pay invoice from outgoing mint
        quote = await outgoing_wallet.melt_quote(invoice.bolt11)
        total_amount = quote.amount + quote.fee_reserve
        if outgoing_wallet.available_balance < total_amount:
            raise Exception("balance too low")
        send_proofs, fees = await outgoing_wallet.select_to_send(
            outgoing_wallet.proofs, total_amount, set_reserved=True
        )

        try:
            await outgoing_wallet.melt(
                proofs=send_proofs, invoice=invoice.bolt11, fee_reserve_sat=quote.fee_reserve, quote_id=quote.quote
            )
        except:
            print("anyways..")

        # mint token in incoming mint
        await incoming_wallet.mint(amountinsats, id=invoice.id)

        await incoming_wallet.load_proofs(reload=True)
        await print_mint_balances(incoming_wallet, show_mints=True)

    async def set_profile(self, name, about, lud16, image, client, keys):
        metadata = Metadata() \
            .set_name(name) \
            .set_display_name(name) \
            .set_about(about) \
            .set_picture(image) \
            .set_lud16(lud16) \
            .set_nip05("")
        print("[" + name + "] Setting profile metadata for " + keys.public_key().to_bech32() + "...")
        print(metadata.as_json())
        await client.set_metadata(metadata)



import json
import os
import re
import sys
from typing import Any
from urllib.request import urlopen, Request

import bs4
import requests
import yt_dlp
from tqdm import tqdm

browser = "chrome"  # "firefox"


def download_xvideo(url, target_location) -> None:
    response = requests.get(url, stream=True)
    total_size = int(response.headers.get("content-length", 0))
    block_size = 1024
    progress_bar = tqdm(total=total_size, unit="B", unit_scale=True)
    download_path = target_location
    with open(download_path, "wb") as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)

    progress_bar.close()
    print("Video downloaded successfully!")


def XDownload(url, filepath=""):
    api_url = f"https://twitsave.com/info?url={url}"

    response = requests.get(api_url)
    data = bs4.BeautifulSoup(response.text, "html.parser")
    download_button = data.find_all("div", class_="origin-top-right")[0]
    quality_buttons = download_button.find_all("a")
    highest_quality_url = quality_buttons[0].get("href")  # Highest quality video url
    download_xvideo(highest_quality_url, filepath)
    return filepath


def getDict() -> dict:
    response = requests.get('https://ttdownloader.com/')
    point = response.text.find('<input type="hidden" id="token" name="token" value="') + \
            len('<input type="hidden" id="token" name="token" value="')
    token = response.text[point:point + 64]
    TTDict = {
        'token': token,
    }

    for i in response.cookies:
        TTDict[str(i).split()[1].split('=')[0].strip()] = str(
            i).split()[1].split('=')[1].strip()
    return TTDict


def createHeader(parseDict) -> tuple[dict[str, Any], dict[str | Any, str | Any], dict[str, str | Any]]:
    cookies = {
        'PHPSESSID': parseDict['PHPSESSID'],
        # 'popCookie': parseDict['popCookie'],
    }
    headers = {
        'authority': 'ttdownloader.com',
        'accept': '*/*',
        'accept-language': 'en-US,en;q=0.9',
        'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
        'origin': 'https://ttdownloader.com',
        'referer': 'https://ttdownloader.com/',
        'sec-ch-ua': '"Not?A_Brand";v="8", "Chromium";v="108", "Google Chrome";v="108"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                      'Chrome/108.0.0.0 Safari/537.36',
        'x-requested-with': 'XMLHttpRequest',
    }
    data = {
        'url': '',
        'format': '',
        'token': parseDict['token'],
    }
    return cookies, headers, data


def TikTokDownload(cookies, headers, data, name, path) -> str:
    response = requests.post('https://ttdownloader.com/search/',
                             cookies=cookies, headers=headers, data=data)
    parsed_link = [i for i in str(response.text).split()
                   if i.startswith("href=")][0]

    response = requests.get(parsed_link[6:-10])
    with open(path + "\\" + name + ".mp4", "wb") as f:
        f.write(response.content)
    return path + "\\" + name + ".mp4"


def TiktokDownloadAll(linkList, path) -> str:
    parseDict = getDict()
    cookies, headers, data = createHeader(parseDict)
    # linkList = getLinkDict()['tiktok']
    for i in linkList:
        try:
            data['url'] = i
            result = TikTokDownload(cookies, headers, data, "tiktok", path)  # str(linkList.index(i))
            return result
        except IndexError:
            parseDict = getDict()
            cookies, headers, data = createHeader(parseDict)
        except Exception as err:
            print(err)
            exit(1)


def YTDownload(link, path, audio_only=True):
    if audio_only:
        return get_audio([link])
    else:
        return get_video([link])


def get_media_duration(url):
    try:
        # ℹ️ See help(yt_dlp.YoutubeDL) for a list of available options and public functions
        ydl_opts = {
            'cookiesfrombrowser': (browser, None, None, None),
        }
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            info = ydl.extract_info(url, download=False)

            # ℹ️ ydl.sanitize_info makes the info json-serializable
            return float(json.dumps(ydl.sanitize_info(info)["duration"]))
    except:
        return None


def get_media_info(url):
    try:
        # ℹ️ See help(yt_dlp.YoutubeDL) for a list of available options and public functions
        ydl_opts = {
            'cookiesfrombrowser': (browser, None, None, None),
        }
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            info = ydl.extract_info(url, download=False)

            # ℹ️ ydl.sanitize_info makes the info json-serializable
            return json.dumps(ydl.sanitize_info(info))
    except:
        return None


def get_audio(URLS):
    try:
        ydl_opts = {
            'cookiesfrombrowser': (browser, None, None, None),
            'format': 'm4a/bestaudio/best',
            "outtmpl": 'outputs/audio',
            'overwrites': 'True',
            # ℹ️ See help(yt_dlp.postprocessor) for a list of available Postprocessors and their arguments
            'postprocessors': [{  # Extract audio using ffmpeg
                'key': 'FFmpegExtractAudio',
                'preferredcodec': 'mp3',
            }]
        }
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            error_code = ydl.download(URLS)

        return "outputs/audio.mp3"
    except:
        return None


def get_video(URLS):
    try:
        ydl_opts = {
            'cookiesfrombrowser': (browser, None, None, None),
            'format': 'mp4',
            'overwrites': 'True',
            # "outtmpl": '/%(uploader)s_%(title)s.%(ext)s',
            "outtmpl": 'outputs/video.mp4',
        }

        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            ydl.download(URLS)
        return "outputs/video.mp4"

    except:
        return None


# OVERCAST
def OvercastDownload(source_url, target_location):
    def get_title(html_str):
        """Get the title from the meta tags"""

        title = re.findall(r"<meta name=\"og:title\" content=\"(.+)\"", html_str)
        if len(title) == 1:
            return title[0].replace("&mdash;", "-")
        return None

    def get_description(html_str):
        """Get the description from the Meta tag"""

        desc_re = r"<meta name=\"og:description\" content=\"(.+)\""
        description = re.findall(desc_re, html_str)
        if len(description) == 1:
            return description[0]
        return None

    def get_url(html_string):
        """Find the URL from the <audio><relay_timeout>.... tag"""

        url = re.findall(r"<relay_timeout src=\"(.+?)\"", html_string)
        if len(url) == 1:
            # strip off the last 4 characters to cater for the #t=0 in the URL
            # which urlretrieve flags as invalid
            return url[0][:-4]
        return None

    """Given a Overcast relay_timeout URL fetch the file it points to"""
    headers = {
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) "
                      "AppleWebKit/537.11 (KHTML, like Gecko) "
                      "Chrome/23.0.1271.64 Safari/537.11",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
        "Accept-Charset": "ISO-8859-1,utf-8;q=0.7,*;q=0.3",
        "Accept-Encoding": "none",
        "Accept-Language": "en-US,en;q=0.8",
        "Connection": "keep-alive",
    }
    req = Request(source_url, None, headers)
    source_data = urlopen(req).read().decode('utf-8')
    title = get_title(source_data)
    url = get_url(source_data)

    if url is None or title is None:
        sys.exit("Could not find parse URL")
    if not os.path.exists(target_location):
        req = requests.get(url)
        file = open(target_location, 'wb')
        for chunk in req.iter_content(100000):
            file.write(chunk)
        file.close()



import json
import os
import re

from nostr_sdk import Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag
from nostr_dvm.utils.nostr_utils import get_event_by_id

"""
This File contains a Module to extract Text from a PDF file locally on the DVM Machine

Accepted Inputs: Url to pdf file, Event containing an URL to a PDF file
Outputs: Text containing the extracted contents of the PDF file
Params:  None
"""


class TextExtractionPDF(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_EXTRACT_TEXT
    TASK: str = "pdf-to-text"
    FIX_COST: float = 0
    dependencies = [("nostr-dvm", "nostr-dvm"),
                    ("pypdf", "pypdf==3.17.1")]

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "url" and input_type != "event":
                    return False
        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex()}

        # default values
        input_type = "url"
        input_content = ""
        url = ""

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                input_content = tag.as_vec()[1]

        if input_type == "url":
            url = input_content
        # if event contains url to pdf, we checked for a pdf link before
        elif input_type == "event":
            evt = await get_event_by_id(input_content, client=client, config=dvm_config)
            url = re.search("(?P<url>https?://[^\\s]+)", evt.content()).group("url")

        options = {
            "url": url,
        }
        request_form['options'] = json.dumps(options)
        return request_form

    async def process(self, request_form):
        from pypdf import PdfReader
        from pathlib import Path
        import requests

        options = self.set_options(request_form)

        try:
            file_path = Path('temp.pdf')
            response = requests.get(options["url"])
            file_path.write_bytes(response.content)
            reader = PdfReader(file_path)
            number_of_pages = len(reader.pages)
            text = ""
            for page_num in range(number_of_pages):
                page = reader.pages[page_num]
                text = text + page.extract_text()

            os.remove('temp.pdf')
            return text
        except Exception as e:
            raise Exception(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    admin_config.LUD16 = dvm_config.LN_ADDRESS
    # Add NIP89
    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/c33ca6fc4cc038ca4adb46fdfdfda34951656f87ee364ef59095bae1495ce669.jpg",
        "about": "I extract text from pdf documents. I only support Latin letters",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {}
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)
    return TextExtractionPDF(name=name, dvm_config=dvm_config, nip89config=nip89config,
                             admin_config=admin_config)


if __name__ == '__main__':
    process_venv(TextExtractionPDF)



import json
import os
from multiprocessing.pool import ThreadPool

from nostr_sdk import Kind

from nostr_dvm.backends.nova_server.utils import check_server_status, send_request_to_server
from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag

"""
This File contains a Module to upscale an image from an url by factor 2-4 

Accepted Inputs: link to image (url)
Outputs: An url to an Image
Params: -upscale        2,3,4
"""


class ImageUpscale(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_GENERATE_IMAGE
    TASK: str = "image-to-image"
    FIX_COST: float = 20

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        hasurl = False
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type == "url":
                    hasurl = True

        if not hasurl:
            return False

        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex() + "_" + self.NAME.replace(" ", "")}
        request_form["trainerFilePath"] = r'modules\image_upscale\image_upscale_realesrgan.trainer'
        url = ""
        out_scale = 4

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "url":
                    url = tag.as_vec()[1]

            elif tag.as_vec()[0] == 'param':
                print("Param: " + tag.as_vec()[1] + ": " + tag.as_vec()[2])
                if tag.as_vec()[1] == "upscale":
                    out_scale = int(tag.as_vec()[2])

        io_input_image = {
            "id": "input_image",
            "type": "input",
            "src": "url:image",
            "uri": url
        }

        io_output = {
            "id": "output_image",
            "type": "output",
            "src": "request:image"
        }

        request_form['data'] = json.dumps([io_input_image, io_output])

        options = {"model": "RealESRGAN_x4plus",
                   "outscale": out_scale,
                   "denoise_strength": 0.5,
                   "tile": 0,
                   "tile_pad": 10,
                   "pre_pad": 0,
                   "compute_type": "fp32",
                   "face_enhance": False}

        request_form['options'] = json.dumps(options)

        return request_form

    async def process(self, request_form):
        try:
            # Call the process route of NOVA-Server with our request form.
            response = send_request_to_server(request_form, self.options['server'])
            if bool(json.loads(response)['success']):
                print("Job " + request_form['jobID'] + " sent to server")

            pool = ThreadPool(processes=1)
            thread = pool.apply_async(check_server_status, (request_form['jobID'], self.options['server']))
            print("Wait for results of server...")
            result = thread.get()
            return result

        except Exception as e:
            raise Exception(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config, server_address):
    dvm_config = build_default_config(identifier)
    dvm_config.USE_OWN_VENV = False
    admin_config.LUD16 = dvm_config.LN_ADDRESS

    # A module might have options it can be initialized with, here we set a default model, lora and the server
    options = {'server': server_address}

    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/229c14e440895da30de77b3ca145d66d4b04efb4027ba3c44ca147eecde891f1.jpg",
        "about": "I upscale an image using realESRGan up to factor 4 (default is factor 4)",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {
            "upscale": {
                "required": False,
                "values": ["2", "3", "4"]
            }
        }
    }
    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return ImageUpscale(name=name, dvm_config=dvm_config, nip89config=nip89config,
                        admin_config=admin_config, options=options)


if __name__ == '__main__':
    process_venv(ImageUpscale)



from nostr_sdk import Keys, EventBuilder, Kind

from nostr_dvm.utils.nostr_utils import send_event
from nostr_dvm.utils.print_utils import bcolors


async def gallery_announce_list(tags, dvm_config, client):
    keys = Keys.parse(dvm_config.NIP89.PK)
    content = ""
    event = EventBuilder(Kind(10011), content, tags).to_event(keys)
    eventid = await send_event(event, client=client, dvm_config=dvm_config, blastr=True)

    print(
        bcolors.BLUE + "[" + dvm_config.NIP89.NAME + "] Announced Gallery for " + dvm_config.NIP89.NAME + " (EventID: " + str(
            eventid.to_hex()) + ")" + bcolors.ENDC)



import asyncio
import json
import time
from pathlib import Path
from threading import Thread

from nostr_dvm.utils.nip65_utils import nip65_announce_relays
from nostr_dvm.utils.nut_wallet_utils import NutZapWallet
from nostr_dvm.utils.print_utils import bcolors

import dotenv
from nostr_sdk import Keys, Client, Tag, EventBuilder, Filter, HandleNotification, Timestamp, nip04_decrypt, \
    nip04_encrypt, NostrSigner, PublicKey, Event, Kind, RelayOptions

from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nostr_utils import send_event, check_and_set_private_key
from nostr_dvm.utils.definitions import EventDefinitions


# TODO HINT: Best use this path with a previously whitelisted privkey, as zapping events is not implemented in the lib/code
async def nostr_client_test_translation(input, kind, lang, sats, satsmax):
    keys = Keys.parse(check_and_set_private_key("test_client"))
    if kind == "text":
        iTag = Tag.parse(["i", input, "text"])
    elif kind == "event":
        iTag = Tag.parse(["i", input, "event"])
    paramTag1 = Tag.parse(["param", "language", lang])

    bidTag = Tag.parse(['bid', str(sats * 1000), str(satsmax * 1000)])
    relaysTag = Tag.parse(['relays', "wss://relay.damus.io", "wss://blastr.f7z.xyz", "wss://relayable.org",
                           "wss://nostr-pub.wellorder.net"])
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to translate a given Input"])
    event = EventBuilder(EventDefinitions.KIND_NIP90_TRANSLATE_TEXT, str("Translate the given input."),
                         [iTag, paramTag1, bidTag, relaysTag, alttag]).to_event(keys)

    relay_list = ["wss://relay.damus.io", "wss://blastr.f7z.xyz", "wss://relayable.org",
                  "wss://nostr-pub.wellorder.net"]

    signer = NostrSigner.keys(keys)
    client = Client(signer)

    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_test_search_profile(input):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    iTag = Tag.parse(["i", input, "text"])
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to translate a given Input"])
    event = EventBuilder(EventDefinitions.KIND_NIP90_USER_SEARCH, str("Search for user"),
                         [iTag, alttag]).to_event(keys)

    relay_list = ["wss://relay.damus.io", "wss://blastr.f7z.xyz", "wss://relayable.org",
                  "wss://nostr-pub.wellorder.net"]

    signer = NostrSigner.keys(keys)
    client = Client(signer)

    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_test_image(prompt):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    iTag = Tag.parse(["i", prompt, "text"])
    outTag = Tag.parse(["output", "image/png;format=url"])
    paramTag1 = Tag.parse(["param", "size", "1024x1024"])

    bidTag = Tag.parse(['bid', str(1000 * 1000), str(1000 * 1000)])
    relaysTag = Tag.parse(['relays', "wss://relay.primal.net", "wss://nostr.oxtr.dev"])
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to generate an Image from a given Input"])
    event = EventBuilder(EventDefinitions.KIND_NIP90_GENERATE_IMAGE, str("Generate an Image."),
                         [iTag, outTag, paramTag1, bidTag, relaysTag, alttag]).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in DVMConfig().RELAY_LIST:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    #config.NIP89.PK = keys.secret_key().to_hex()
    #await nip65_announce_relays(config, client=client)
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_test_censor_filter(users):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://relay.damus.io", "wss://blastr.f7z.xyz", "wss://relayable.org",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to find people to ignore based on people the user trusts"])
    # pTag = Tag.parse(["p", user, "text"])
    tags = [relaysTag, alttag]
    for user in users:
        iTag = Tag.parse(["i", user, "text"])
        tags.append(iTag)

    event = EventBuilder(EventDefinitions.KIND_NIP90_PEOPLE_DISCOVERY, str("Give me bad actors"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_test_inactive_filter(user):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://relay.damus.io", "wss://blastr.f7z.xyz",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to find people that are inactive"])
    paramTag = Tag.parse(["param", "user", user])
    paramTag2 = Tag.parse(["param", "since_days", "120"])

    tags = [relaysTag, alttag, paramTag, paramTag2]

    event = EventBuilder(EventDefinitions.KIND_NIP90_PEOPLE_DISCOVERY, str("Give me inactive users"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    await client.add_relay("wss://nostr.band")
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_test_tts(prompt):
    keys = Keys.parse(check_and_set_private_key("test_client"))
    iTag = Tag.parse(["i", "9d867cd3e868111a31c8acfa41ab7523b9940fc46c804d7db89d7f373c007fa6", "event"])
    # iTag = Tag.parse(["i", prompt, "text"])
    paramTag1 = Tag.parse(["param", "language", "en"])

    bidTag = Tag.parse(['bid', str(1000 * 1000), str(1000 * 1000)])
    relaysTag = Tag.parse(['relays', "wss://relay.damus.io", "wss://blastr.f7z.xyz", "wss://relayable.org",
                           "wss://nostr-pub.wellorder.net"])
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to generate TTSt"])
    event = EventBuilder(EventDefinitions.KIND_NIP90_TEXT_TO_SPEECH, str("Generate an Audio File."),
                         [iTag, paramTag1, bidTag, relaysTag, alttag]).to_event(keys)

    relay_list = ["wss://relay.damus.io", "wss://blastr.f7z.xyz", "wss://relayable.org" "wss://dvms.f7z.io",
                  ]

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_test_discovery(user, ptag):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://relay.damus.io", "wss://blastr.f7z.xyz",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to find content"])
    paramTag = Tag.parse(["param", "user", user])
    pTag = Tag.parse(["p", ptag])

    tags = [relaysTag, alttag, paramTag, pTag]

    event = EventBuilder(EventDefinitions.KIND_NIP90_CONTENT_DISCOVERY, str("Give me content"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    ropts = RelayOptions().ping(False)
    await client.add_relay("wss://nostr.band")
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()



async def nostr_client_custom_discovery(user, ptag):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://nostr.oxtr.dev", "wss://relay.primal.net",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to find content"])
    paramTag = Tag.parse(["param", "user", user])

    search = " art , photograph , photo , photography , painting ,#artstr, drawing "
    avoid = "sex"
    must = "http"

    paramTagSearch = Tag.parse(["param", "search_list", search])
    paramTagAvoid = Tag.parse(["param", "avoid_list", avoid])
    paramTagMust = Tag.parse(["param", "must_list", must])

    pTag = Tag.parse(["p", ptag])

    tags = [relaysTag, alttag, paramTag, pTag]# paramTagSearch, paramTagMust, paramTagAvoid]

    event = EventBuilder(EventDefinitions.KIND_NIP90_CONTENT_DISCOVERY, str("Give me content"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    ropts = RelayOptions().ping(False)

    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_generic_test(ptag):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://nostr.oxtr.dev", "wss://relay.primal.net",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task"])

    pTag = Tag.parse(["p", ptag])

    tags = [relaysTag, alttag, pTag]

    event = EventBuilder(Kind(5050), str("Give me content"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    ropts = RelayOptions().ping(False)
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_duckduck_test(ptag, query):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://nostr.oxtr.dev", "wss://relay.primal.net",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task"])

    pTag = Tag.parse(["p", ptag])
    iTag = Tag.parse(["i", query, "text"])

    tags = [relaysTag, alttag, pTag, iTag]

    event = EventBuilder(Kind(5050), str("Give me content"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    ropts = RelayOptions().ping(False)
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()
async def nostr_client_flux_schnell(ptag, query):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://nostr.oxtr.dev", "wss://relay.primal.net",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task"])

    pTag = Tag.parse(["p", ptag])
    iTag = Tag.parse(["i", query, "text"])

    tags = [relaysTag, alttag, pTag, iTag]

    event = EventBuilder(Kind(5100), str("Give me image"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    ropts = RelayOptions().ping(False)
    await client.connect()
    config = DVMConfig
    await send_event(event, client=client, dvm_config=config)
    return event.as_json()




async def nostr_client_test_discovery_user(user, ptag):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://relay.primal.net", "wss://dvms.f7z.io",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to find users"])
    paramTag = Tag.parse(["param", "user", user])
    pTag = Tag.parse(["p", ptag])

    tags = [relaysTag, alttag, paramTag, pTag]

    event = EventBuilder(EventDefinitions.KIND_NIP90_PEOPLE_DISCOVERY, str("Give me people"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    eventid = await send_event(event, client=client, dvm_config=config)
    return event.as_json()


async def nostr_client_test_discovery_gallery(user, ptag):
    keys = Keys.parse(check_and_set_private_key("test_client"))

    relay_list = ["wss://relay.damus.io", "wss://dvms.f7z.io",
                  ]

    relaysTag = Tag.parse(relay_list)
    alttag = Tag.parse(["alt", "This is a NIP90 DVM AI task to find users"])
    paramTag = Tag.parse(["param", "user", user])
    pTag = Tag.parse(["p", ptag])

    tags = [relaysTag, alttag, paramTag, pTag]

    event = EventBuilder(EventDefinitions.KIND_NIP90_VISUAL_DISCOVERY, str("Give me visuals"),
                         tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    eventid = await send_event(event, client=client, dvm_config=config)
    print(eventid.to_hex())
    return event.as_json()


async def nostr_client_test_image_private(prompt, cashutoken):
    keys = Keys.parse(check_and_set_private_key("test_client"))
    receiver_keys = Keys.parse(check_and_set_private_key("replicate_sdxl"))

    relay_list = ["wss://relay.damus.io", "wss://blastr.f7z.xyz", "wss://relayable.org",
                  "wss://nostr-pub.wellorder.net"]
    i_tag = Tag.parse(["i", prompt, "text"])
    outTag = Tag.parse(["output", "image/png;format=url"])
    paramTag1 = Tag.parse(["param", "size", "1024x1024"])
    pTag = Tag.parse(["p", receiver_keys.public_key().to_hex()])

    bid = str(50 * 1000)
    bid_tag = Tag.parse(['bid', bid, bid])
    relays_tag = Tag.parse(["relays", json.dumps(relay_list)])
    alt_tag = Tag.parse(["alt", "Super secret test"])
    cashu_tag = Tag.parse(["cashu", cashutoken])

    encrypted_params_string = json.dumps([i_tag.as_vec(), outTag.as_vec(), paramTag1.as_vec(), bid_tag.as_vec(),
                                          relays_tag.as_vec(), alt_tag.as_vec(), cashu_tag.as_vec()])

    encrypted_params = nip04_encrypt(keys.secret_key(), receiver_keys.public_key(),
                                     encrypted_params_string)

    encrypted_tag = Tag.parse(['encrypted'])
    nip90request = EventBuilder(EventDefinitions.KIND_NIP90_GENERATE_IMAGE, encrypted_params,
                                [pTag, encrypted_tag]).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    config = DVMConfig
    await send_event(nip90request, client=client, dvm_config=config)
    return nip90request.as_json()


async def nostr_client():
    keys = Keys.parse(check_and_set_private_key("test_client"))
    sk = keys.secret_key()
    pk = keys.public_key()
    print(f"Nostr Client public key: {pk.to_bech32()}, Hex: {pk.to_hex()} ")
    signer = NostrSigner.keys(keys)
    client = Client(signer)

    dvmconfig = DVMConfig()
    for relay in dvmconfig.RELAY_LIST:
        await client.add_relay(relay)
    await client.connect()

    dm_zap_filter = Filter().pubkey(pk).kinds([EventDefinitions.KIND_DM,
                                               EventDefinitions.KIND_ZAP]).since(
        Timestamp.now())  # events to us specific
    kinds = [EventDefinitions.KIND_NIP90_GENERIC]
    #SUPPORTED_KINDS = [Kind(6100), Kind(7000)]

    for kind in range(6000, 7001):
        if kind not in kinds:
            kinds.append(Kind(kind))
    dvm_filter = (Filter().kinds(kinds).since(Timestamp.now()).pubkey(pk))
    await client.subscribe([dm_zap_filter, dvm_filter], None)

    # await nostr_client_test_translation("This is the result of the DVM in spanish", "text", "es", 20, 20)
    # await nostr_client_test_translation("note1p8cx2dz5ss5gnk7c59zjydcncx6a754c0hsyakjvnw8xwlm5hymsnc23rs", "event", "es", 20,20)
    # await nostr_client_test_translation("44a0a8b395ade39d46b9d20038b3f0c8a11168e67c442e3ece95e4a1703e2beb", "event", "zh", 20, 20)

    #await nostr_client_test_image("a beautiful purple ostrich watching the sunset, eating a cashew nut")
    await nostr_client_custom_discovery("99bb5591c9116600f845107d31f9b59e2f7c7e09a1ff802e84f1d43da557ca64", "7240284b84951cfedbc20fce26f0e3f0a36da3e9c1be85d7a06965f0d4fe25fb")
    #"a018ba05af400b52772e33162d8326fca4a167fe7b6d3cd2382e14cac2af6841"
   # await nostr_client_duckduck_test(PublicKey.parse("7a63849b684d90c0de983492578b12e147e56f5d79ed6585cc64e5aa8a122744").to_hex() , "How do i create a dockerfile for python 3.12")
    #await nostr_client_flux_schnell("d57f1efb7582f58cade6f482d53eefa998d8082711b996aae3dc5f5527cbdd6e" , "topics")

    # await nostr_client_test_search_profile("dontbelieve")
    #wot = ["99bb5591c9116600f845107d31f9b59e2f7c7e09a1ff802e84f1d43da557ca64"]
    #await nostr_client_test_discovery("99bb5591c9116600f845107d31f9b59e2f7c7e09a1ff802e84f1d43da557ca64", "ab6cdf12ca3ae5109416295b8cd8a53fdec3a9d54beb7a9aee0ebfb67cb4edf7")
    # await nostr_client_test_discovery_gallery("99bb5591c9116600f845107d31f9b59e2f7c7e09a1ff802e84f1d43da557ca64", "4add3944eb596a27a650f9b954f5ed8dfefeec6ca50473605b0fbb058dd11306")

    #await nostr_client_test_discovery("99bb5591c9116600f845107d31f9b59e2f7c7e09a1ff802e84f1d43da557ca64",
    #                                      "7a63849b684d90c0de983492578b12e147e56f5d79ed6585cc64e5aa8a122744")

    # await nostr_client_test_censor_filter(wot)
    # await nostr_client_test_inactive_filter("99bb5591c9116600f845107d31f9b59e2f7c7e09a1ff802e84f1d43da557ca64")

    #await nostr_client_test_tts("Hello, this is a test. Mic check one, two.")

    # cashutoken = "cashuAeyJ0b2tlbiI6W3sicHJvb2ZzIjpbeyJpZCI6InZxc1VRSVorb0sxOSIsImFtb3VudCI6MSwiQyI6IjAyNWU3ODZhOGFkMmExYTg0N2YxMzNiNGRhM2VhMGIyYWRhZGFkOTRiYzA4M2E2NWJjYjFlOTgwYTE1NGIyMDA2NCIsInNlY3JldCI6InQ1WnphMTZKMGY4UElQZ2FKTEg4V3pPck5rUjhESWhGa291LzVzZFd4S0U9In0seyJpZCI6InZxc1VRSVorb0sxOSIsImFtb3VudCI6NCwiQyI6IjAyOTQxNmZmMTY2MzU5ZWY5ZDc3MDc2MGNjZmY0YzliNTMzMzVmZTA2ZGI5YjBiZDg2Njg5Y2ZiZTIzMjVhYWUwYiIsInNlY3JldCI6IlRPNHB5WE43WlZqaFRQbnBkQ1BldWhncm44UHdUdE5WRUNYWk9MTzZtQXM9In0seyJpZCI6InZxc1VRSVorb0sxOSIsImFtb3VudCI6MTYsIkMiOiIwMmRiZTA3ZjgwYmMzNzE0N2YyMDJkNTZiMGI3ZTIzZTdiNWNkYTBhNmI3Yjg3NDExZWYyOGRiZDg2NjAzNzBlMWIiLCJzZWNyZXQiOiJHYUNIdHhzeG9HM3J2WWNCc0N3V0YxbU1NVXczK0dDN1RKRnVwOHg1cURzPSJ9XSwibWludCI6Imh0dHBzOi8vbG5iaXRzLmJpdGNvaW5maXhlc3RoaXMub3JnL2Nhc2h1L2FwaS92MS9ScDlXZGdKZjlxck51a3M1eVQ2SG5rIn1dfQ=="
    # await nostr_client_test_image_private("a beautiful ostrich watching the sunset")

    nutzap_wallet = NutZapWallet()


    nut_wallet = await nutzap_wallet.get_nut_wallet(client, keys)
    #dangerous, dont use this, except your wallet is messed up.
    delete = False
    if delete:
        for mint in nut_wallet.nutmints:
            await nutzap_wallet.update_spend_mint_proof_event(nut_wallet, mint.proofs, mint.mint_url, "", None,
                                                     None, client, keys)

        nut_wallet.balance = 0
        await nutzap_wallet.update_nut_wallet(nut_wallet, [], client, keys)
        nut_wallet = await nutzap_wallet.get_nut_wallet(client, keys)



    class NotificationHandler(HandleNotification):
        last_event_time = 0
        async def handle(self, relay_url, subscription_id, event: Event):

            print(
                bcolors.BLUE + f"Received new event from {relay_url}: {event.as_json()}" + bcolors.ENDC)
            if event.kind().as_u16() == 7000:
                print("[Nostr Client]: " + event.as_json())
                amount_sats = 0
                status = ""
                for tag in event.tags():
                    if tag.as_vec()[0] == "amount":
                        amount_sats = int(int(tag.as_vec()[1]) / 1000) # millisats
                    if tag.as_vec()[0] == "status":
                       status = tag.as_vec()[1]
                # THIS IS FOR TESTING
                if event.author().to_hex() == "89669b03bb25232f33192fdda77b8e36e3d3886e9b55b3c74b95091e916c8f98" and status == "payment-required" and event.created_at().as_secs() > self.last_event_time:
                    self.last_event_time = event.created_at().as_secs()
                    nut_wallet = await nutzap_wallet.get_nut_wallet(client, keys)
                    if nut_wallet is None:
                        await nutzap_wallet.create_new_nut_wallet(dvmconfig.NUZAP_MINTS, dvmconfig.NUTZAP_RELAYS, client, keys, "Test", "My Nutsack")
                        nut_wallet = await nutzap_wallet.get_nut_wallet(client, keys)
                        if nut_wallet is not None:
                            await nutzap_wallet.announce_nutzap_info_event(nut_wallet, client, keys)
                        else:
                            print("Couldn't fetch wallet, please restart and see if it is there")

                    await nutzap_wallet.send_nut_zap(amount_sats, "From my nutsack lol", nut_wallet, event.id().to_hex(),
                                                     event.author().to_hex(), client,
                                                     keys)


            elif 6000 < event.kind().as_u16() < 6999:
                print("[Nostr Client]: " + event.as_json())
                print("[Nostr Client]: " + event.content())

            elif event.kind().as_u16() == 4:
                dec_text = nip04_decrypt(sk, event.author(), event.content())
                print("[Nostr Client]: " + f"Received new msg: {dec_text}")

            elif event.kind().as_u16() == 9735:
                print("[Nostr Client]: " + f"Received new zap:")
                print(event.as_json())

        async def handle_msg(self, relay_url, msg):
            return

    asyncio.create_task(client.handle_notifications(NotificationHandler()))
    # await client.handle_notifications(NotificationHandler())
    while True:
        await asyncio.sleep(2)


if __name__ == '__main__':

    env_path = Path('.env')
    if env_path.is_file():
        print(f'loading environment from {env_path.resolve()}')
        dotenv.load_dotenv(env_path, verbose=True, override=True)
    else:
        raise FileNotFoundError(f'.env file not found at {env_path} ')

    asyncio.run(nostr_client())



import json
import os

from nostr_sdk import Kind

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface, process_venv
from nostr_dvm.utils.admin_utils import AdminConfig
from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.dvmconfig import DVMConfig, build_default_config
from nostr_dvm.utils.nip88_utils import NIP88Config
from nostr_dvm.utils.nip89_utils import NIP89Config, check_and_set_d_tag

"""
This File contains a Module to generate Text, based on a prompt using a the HuggingChat LLM on Huggingface

Accepted Inputs: Prompt (text)
Outputs: Generated text
"""


class TextGenerationHuggingChat(DVMTaskInterface):
    KIND: Kind = EventDefinitions.KIND_NIP90_GENERATE_TEXT
    TASK: str = "text-to-text"
    FIX_COST: float = 0
    dependencies = [("nostr-dvm", "nostr-dvm"),
                    ("hugchat", "hugchat")]

    async def init_dvm(self, name, dvm_config: DVMConfig, nip89config: NIP89Config, nip88config: NIP88Config = None,
                       admin_config: AdminConfig = None, options=None):
        dvm_config.SCRIPT = os.path.abspath(__file__)

    async def is_input_supported(self, tags, client=None, dvm_config=None):
        for tag in tags:
            if tag.as_vec()[0] == 'i':
                input_value = tag.as_vec()[1]
                input_type = tag.as_vec()[2]
                if input_type != "text":
                    return False

        return True

    async def create_request_from_nostr_event(self, event, client=None, dvm_config=None):
        request_form = {"jobID": event.id().to_hex() + "_" + self.NAME.replace(" ", "")}
        prompt = ""

        for tag in event.tags():
            if tag.as_vec()[0] == 'i':
                input_type = tag.as_vec()[2]
                if input_type == "text":
                    prompt = tag.as_vec()[1]

        options = {
            "prompt": prompt,
        }
        request_form['options'] = json.dumps(options)

        return request_form

    async def process(self, request_form):
        from hugchat import hugchat
        from hugchat.login import Login
        sign = Login(os.getenv("HUGGINGFACE_EMAIL"), os.getenv("HUGGINGFACE_PASSWORD"))
        cookie_path_dir = "./cookies_snapshot"
        try:
            cookies = sign.loadCookiesFromDir(
                cookie_path_dir)  # This will detect if the JSON file exists, return cookies if it does and raise an Exception if it's not.
        except:
            cookies = sign.login()
            sign.saveCookiesToDir(cookie_path_dir)

        options = self.set_options(request_form)

        try:
            chatbot = hugchat.ChatBot(cookies=cookies.get_dict())  # or cookie_path="usercookies/<email>.json"
            query_result = chatbot.query(options["prompt"])
            print(query_result["text"])  # or query_result.text or query_result["text"]
            return str(query_result["text"]).lstrip()

        except Exception as e:
            print("Error in Module: " + str(e))
            raise Exception(e)


# We build an example here that we can call by either calling this file directly from the main directory,
# or by adding it to our playground. You can call the example and adjust it to your needs or redefine it in the
# playground or elsewhere
def build_example(name, identifier, admin_config):
    dvm_config = build_default_config(identifier)
    admin_config.LUD16 = dvm_config.LN_ADDRESS

    nip89info = {
        "name": name,
        "image": "https://image.nostr.build/720eadc9af89084bb09de659af43ad17fec1f4b0887084e83ac0ae708dfa83a6.png",
        "about": "I use a LLM connected via Huggingchat",
        "encryptionSupported": True,
        "cashuAccepted": True,
        "nip90Params": {}
    }

    nip89config = NIP89Config()
    nip89config.DTAG = check_and_set_d_tag(identifier, name, dvm_config.PRIVATE_KEY, nip89info["image"])
    nip89config.CONTENT = json.dumps(nip89info)

    return TextGenerationHuggingChat(name=name, dvm_config=dvm_config, nip89config=nip89config,
                                     admin_config=admin_config)


if __name__ == '__main__':
    process_venv(TextGenerationHuggingChat)



import asyncio
import datetime
# General
import json
import random
import time
from itertools import islice

import networkx as nx
import nostr_sdk
import numpy as np
from nostr_sdk import Options, Keys, NostrSigner, ClientBuilder, Kind, PublicKey, Filter
from scipy.sparse import lil_matrix, isspmatrix_csr

from nostr_dvm.utils.definitions import relay_timeout
from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nostr_utils import check_and_set_private_key


async def get_following(pks, max_time_request=10, newer_than_time=None):
    '''
        OUTPUT: following; a networkx graph
        > each node has associated the timestamp of the latest retrivable kind3 event

        > it's possible to limit the search to events newer_than_time

        NOTE: do not get_following of more than 1000 pks;
        instead, divide the request into batches of smaller size (500 recommended)
    '''

    # handling the case of a single key passed
    if type(pks) == str:
        pks = [pks]

    # transforming into PubKey object type
    list_pk = [nostr_sdk.PublicKey.parse(pk) for pk in pks]

    # newer_than_time provided? If so, it only fetch events that are newer
    if newer_than_time is None:
        filter = nostr_sdk.Filter().authors(list_pk).kind(Kind(3))

    else:
        newer_than_time = round(newer_than_time)
        ts = nostr_sdk.Timestamp().from_secs(newer_than_time)
        filter = nostr_sdk.Filter().authors(list_pk).kind(3).since(ts)

    # fetching events
    opts = (Options().wait_for_send(False).send_timeout(datetime.timedelta(seconds=5)))
    keys = Keys.parse(check_and_set_private_key("test_client"))
    signer = NostrSigner.keys(keys)
    cli = ClientBuilder().signer(signer).opts(opts).build()

    for relay in DVMConfig.RECONCILE_DB_RELAY_LIST:
        await cli.add_relay(relay)

    await cli.connect()

    events = await cli.get_events_of([filter], relay_timeout)

    # initializing the graph structure
    following = nx.DiGraph()
    following.add_nodes_from(pks)

    if events == []:
        return following

    for event in events:

        author = event.author().to_hex()

        # events are returned based on the timestamp,
        # the first event for each pubkey is the most recent = the one we should use

        if event.verify() and author in following.nodes() and 'timestamp' not in following.nodes[author]:
            # updating the nodes and edges
            nodes = event.public_keys()

            # converting to hex and removing self-following
            nodes = [pk.to_hex() for pk in nodes if pk.to_hex() != author]

            following.update(edges=[(author, pk) for pk in nodes], nodes=nodes)

            # updating the timestamp
            tp = event.created_at().as_secs()
            following.nodes[author]['timestamp'] = tp

    return following


async def build_wot_network(seed_pks, depth=2, max_batch=500, max_time_request=10):
    if not seed_pks:
        print('Error: seed_pks cannot be empty')
        return

    if depth < 1:
        print('Error: depth cannot be lower than 1')
        return

        # handling the case of a single key being passed
    if type(seed_pks) == str:
        seed_pks = [seed_pks]

    print('Step 1: fetching kind 3 events from relays & pre-processing')

    tic = time.time()

    # initialize the index_map
    index_map = {pk: i for i, pk in enumerate(seed_pks)}

    # initialize the graph
    seed_graph = nx.DiGraph()
    seed_graph.add_nodes_from(index_map.values())

    # build the network internal function
    index_map, network_graph = await _build_network_from(index_map, seed_graph, set(), depth, max_batch,
                                                         max_time_request)

    toc = time.time()

    print('\nFinished in ' + str(toc - tic))

    return index_map, network_graph


async def _build_network_from(index_map, network_graph, visited_pk=None, depth=2, max_batch=500, max_time_request=10):
    '''
        OUTPUTS:

        1. index_map = {pk0 : 0, pk1 : 1, ... }
        > an index that translates pub keys in hex format into nodes in the graph

        2. network_graph; a networkx directed graph about who follows who
        > each node has associated the timestamp of the latest retrivable kind3 event
        > if the node is a leaf or doesn't follow anyone, it has no attribute 'timestamp'

    '''

    # pks to be visited next, splitted in batches
    if visited_pk is None:
        visited_pk = set()
    to_visit_pk = split_set(index_map.keys() - visited_pk, max_batch)

    for pks in to_visit_pk:
        # getting the followings as a graph
        following = await get_following(pks, max_time_request)

        # update the visited_pk
        visited_pk.update(pks)

        # add the new pub_keys to the index_map
        index_map = _extend_index_map(index_map, following)

        # re-lable nodes using the index_map to use storage more efficiently
        nx.relabel_nodes(following, index_map, copy=False)

        # extend the network graph
        network_graph.update(following)

    if depth == 1:
        return index_map, network_graph

    else:

        # recursive call
        index_map, network_graph = await _build_network_from(index_map, network_graph, visited_pk, depth - 1, max_batch,
                                                             max_time_request)

    print('current network: ' + str(len(network_graph.nodes())) + ' npubs', end='\r')

    return index_map, network_graph


def _extend_index_map(index_map, following):
    '''
        index_map = { pk0: 0, pk1:1, ...}
    '''

    # getting all new pubkeys
    new_pk = following.nodes() - index_map.keys()

    # start assigning new indices from the next available index
    start_index = len(index_map)

    # Update the index_map with new keys and their assigned indices
    for i, pk in enumerate(new_pk, start=start_index):
        index_map[pk] = i

    return index_map


def split_set(my_set, max_batch):
    my_list = list(my_set)
    return [set(my_list[x: x + max_batch]) for x in range(0, len(my_set), max_batch)]


def save_network(index_map, network_graph, name=None):
    if name == None:
        # adding unix time to file name to avoid replacing an existing file
        name = str(round(time.time()))
    # filename = os.path.join('/cache/', 'index_map_' + name + '.json')
    filename = 'index_map_' + name + '.json'
    # saving the index_map as a json file
    with open(filename, 'w') as f:
        json.dump(index_map, f, indent=4)

    # Convert to node-link format suitable for JSON
    data = nx.node_link_data(network_graph)

    # saving the network_graph as a json file
    # filename = os.path.join('/cache/', 'network_graph_' + name + '.json')
    filename = 'network_graph_' + name + '.json'
    with open(filename, 'w') as f:
        json.dump(data, f)

    print(' > index_map_' + name + '.json')
    print(' > network_graph_' + name + '.json')

    return


def load_network(name):
    if type(name) != str:
        name = str(name)

    # loading the index_map
    with open('index_map_' + name + '.json', 'r') as f:
        index_map = json.load(f)

    # loading the JSON for the graph
    with open('network_graph_' + name + '.json', 'r') as f:
        data = json.load(f)

    # Convert JSON back to graph
    network_graph = nx.node_link_graph(data)

    return index_map, network_graph


def get_mc_pagerank(G, R, nodelist=None, alpha=0.85):
    '''
        Monte-Carlo complete path stopping at dandling nodes

        INPUTS
        ------
        G: graph
            A directed Networkx graph. This function cannot work on directed graphs.

        R: int
            The number of random walks to be performed per node

        nodelist: list, optional
            the list of nodes in G networkx graph.
            It is used to order the nodes in a specified way

        alpha: float, optional
            It is the dampening factor of Pagerank. default value is 0.85

        OUTPUTS
        -------
        walk_visited_count: CSR matrix
            a Compressed Sparse Row (CSR) matrix; element (i,j) is equal to
            the number of times v_j has been visited by a random walk started from v_i

        mc_pagerank: dict
            The dictionary {node: pg} of the pagerank value for each node in G

        References
        ----------
        [1] K.Avrachenkov, N. Litvak, D. Nemirovsky, N. Osipova
        "Monte Carlo methods in PageRank computation: When one iteration is sufficient"
        https://www-sop.inria.fr/members/Konstantin.Avratchenkov/pubs/mc.pdf
    '''

    # validate all the inputs and initialize variables
    N, nodelist, inverse_nodelist = _validate_inputs_and_init_mc(G, R, nodelist, alpha)

    # initialize walk_visited_count as a sparse LIL matrix
    walk_visited_count = lil_matrix((N, N), dtype='int')

    progress_count = 0

    # perform R random walks for each node
    for node in nodelist:

        # print progress every 200 nodes
        progress_count += 1
        if progress_count % 200 == 0:
            print('progress = {:.2f}%'.format(100 * progress_count / N), end='\r')

        for _ in range(R):

            node_pos = inverse_nodelist[node]
            walk_visited_count[node_pos, node_pos] += 1

            current_node = node

            while random.uniform(0, 1) < alpha:

                successors = list(G.successors(current_node))
                if not successors:
                    break

                current_node = random.choice(successors)
                current_node_pos = inverse_nodelist[current_node]

                # add current node to the walk_visited_count
                walk_visited_count[node_pos, current_node_pos] += 1

    # convert lil_matrix to csr_matrix for efficient storage and access
    walk_visited_count = walk_visited_count.tocsr()

    # sum all visits for each node into a numpy array
    total_visited_count = np.array(walk_visited_count.sum(axis=0)).flatten()

    # reciprocal of the number of total visits
    one_over_s = 1 / sum(total_visited_count)

    mc_pagerank = {nodelist[j]: total_visited_count[j] * one_over_s for j in range(N)}

    print('progress = 100%       ', end='\r')
    print('\nTotal walks performed: ', N * R)

    return walk_visited_count, mc_pagerank


def _validate_inputs_and_init_mc(G, R, nodelist, alpha):
    '''
    This function validate the inputs and initialize the following variables:

    N: int
        the number of nodes in G Networkx graph

    nodelist : list
        the list of nodes in G Networkx graph

    inverse_nodelist : dict
       a dictionary that maps each node in G to its position in nodelist
    '''

    N = len(G)
    if N == 0:
        raise ValueError("Graph G is empty")

    if not isinstance(R, int) or R <= 0:
        raise ValueError("R must be a positive integer")

    if not isinstance(alpha, float) or not (0 < alpha < 1):
        raise ValueError("alpha must be a float between 0 and 1")

    if nodelist is not None and set(nodelist) != set(G.nodes()):
        raise ValueError("nodelist does not match the nodes in G")

    elif nodelist is None:
        nodelist = list(G.nodes())

    # compute the inverse map of nodelist
    inverse_nodelist = {nodelist[j]: j for j in range(N)}

    return N, nodelist, inverse_nodelist


def get_subrank(S, G, walk_visited_count, nodelist, alpha=0.85):
    '''
        Subrank algorithm (stopping at dandling nodes);
        it aims to approximate the Pagerank over S subgraph of G

        INPUTS
        ------
        S: graph
            A directed Networkx graph, induced subgraph of G

        G: graph
            A directed Networkx graph. This function cannot work on directed graphs.

        walk_visited_count: CSR matrix
            a Compressed Sparse Row (CSR) matrix; element (i,j) is equal to
            the number of times v_j has been visited by a random walk started from v_i

        nodelist: list, optional
            the list of nodes in G Networkx graph. It is used to decode walk_visited_count

       alpha: float, optional
            It is the dampening factor of Pagerank. default value is 0.85

        OUTPUTS
        -------
        subrank: dict
            The dictionary {node: pg} of the pagerank value for each node in S

        References
        ----------
        [1] Pippellia,
        "Pagerank on subgraphs—efficient Monte-Carlo estimation"
        https://pippellia.com/pippellia/Social+Graph/Pagerank+on+subgraphs%E2%80%94efficient+Monte-Carlo+estimation
    '''

    # validate inputs and initialize variables
    N, S_nodes, G_nodes, inverse_nodelist = _validate_inputs_and_init(S, G, walk_visited_count, nodelist, alpha)

    # compute visited count from walks that started from S
    visited_count_from_S = _get_visited_count_from_S(N, S_nodes, walk_visited_count, nodelist, inverse_nodelist)

    # compute positive and negative walks to do
    positive_walks, negative_walks = _get_walks_to_do(S_nodes, G_nodes, S, G, visited_count_from_S, alpha)

    print(f'walks performed = {sum(positive_walks.values()) + sum(negative_walks.values())}')

    # perform the walks and get the visited counts
    positive_count = _perform_walks(S_nodes, S, positive_walks, alpha)
    negative_count = _perform_walks(S_nodes, S, negative_walks, alpha)

    # add the effects of the random walk to the count of G
    new_visited_count = {node: visited_count_from_S[node] + positive_count[node] - negative_count[node]
                         for node in S_nodes}

    # compute the number of total visits
    total_visits = sum(new_visited_count.values())

    # compute the subrank
    subrank = {node: visits / total_visits
               for node, visits in new_visited_count.items()}

    return subrank


def _validate_inputs_and_init(S, G, walk_visited_count, nodelist, alpha):
    '''
    This function validate the inputs and initialize the following variables:

    N: int
        the number of nodes in G Networkx graph

    S_nodes: set
        the set of nodes that belongs to S

    G_nodes: set
        the set of nodes that belongs to G

    inverse_nodelist : dict
        a dictionary that maps each node in G to its position in nodelist

    Note: S being a subgraph of G is NOT checked because it's computationally expensive.
    '''

    if len(S) == 0:
        raise ValueError("graph S is empty")

    N = len(G)
    if N == 0:
        raise ValueError("graph G is empty")

    if not isinstance(alpha, float) or not (0 < alpha < 1):
        raise ValueError("alpha must be a float between 0 and 1")

    if not isspmatrix_csr(walk_visited_count) or walk_visited_count.shape != (N, N):
        raise ValueError(f"walk_visited_count must be a {(N, N)} CSR matrix")

    S_nodes = set(S.nodes())
    G_nodes = set(G.nodes())

    if not nodelist or set(nodelist) != set(G_nodes):
        raise ValueError("nodelist does not match the nodes in G")

    # compute the inverse map of nodelist
    inverse_nodelist = {nodelist[j]: j for j in range(N)}

    return N, S_nodes, G_nodes, inverse_nodelist


def _get_visited_count_from_S(N, S_nodes, walk_visited_count, nodelist, inverse_nodelist):
    '''
    This function extracts the number of visits that come from walks that started from S
    '''

    # getting the indices of nodes in S
    S_indices = [inverse_nodelist[node] for node in S_nodes]

    # Extract the rows
    S_matrix = walk_visited_count[S_indices, :]

    # Sum the rows
    visited_count_from_S = np.array(S_matrix.sum(axis=0)).flatten()

    # convert to a dictionary
    visited_count_from_S = {nodelist[j]: visited_count_from_S[j] for j in range(N)}

    return visited_count_from_S


def _get_walks_to_do(S_nodes, G_nodes, S, G, visited_count_from_S, alpha):
    '''
    This function calculates the positive and negative walks to be done for each node.
    It is a necessary step to take into account the different structure of S
    with respect to that of G.
    '''

    # compute nodes in G-S
    external_nodes = G_nodes - S_nodes

    # compute nodes in S that point to G-S
    nodes_that_point_externally = {u for u, v in nx.edge_boundary(G, S_nodes, external_nodes)}

    walks_to_do = {node: 0 for node in S_nodes}

    # add positive random walks to walks_to_do
    for node in nodes_that_point_externally:

        successors = set(G.successors(node)) & S_nodes

        if successors:

            # compute estimate visits
            visited_count = visited_count_from_S[node]
            degree_S = S.out_degree(node)
            degree_G = G.out_degree(node)
            estimate_visits = alpha * visited_count * (1 / degree_S - 1 / degree_G)

            for succ in successors:
                walks_to_do[succ] += estimate_visits

    # subtract number of negative random walks
    for node in external_nodes:

        successors = set(G.successors(node)) & S_nodes

        if successors:

            # compute estimate visits
            visited_count = visited_count_from_S[node]
            degree = G.out_degree(node)
            estimate_visits = alpha * visited_count / degree

            for succ in successors:
                walks_to_do[succ] -= estimate_visits

    # split the walks to do into positive and negative
    positive_walks_to_do = {node: round(value) for node, value in walks_to_do.items() if value > 0}
    negative_walks_to_do = {node: round(-value) for node, value in walks_to_do.items() if value < 0}

    return positive_walks_to_do, negative_walks_to_do


def _perform_walks(S_nodes, S, walks_to_do, alpha):
    '''
    This function performs a certain number of random walks on S for each node;
    It then returns the visited count for each node in S.
    '''

    # initializing the visited count
    visited_count = {node: 0 for node in S_nodes}

    for starting_node in walks_to_do.keys():

        num = walks_to_do[starting_node]

        # performing num random walks
        for _ in range(num):

            current_node = starting_node
            visited_count[current_node] += 1

            # performing one random walk
            while random.uniform(0, 1) < alpha:

                successors = list(S.successors(current_node))

                if not successors:
                    break

                current_node = random.choice(successors)

                # updating the visited count
                visited_count[current_node] += 1

    return visited_count


async def get_metadata(npub):
    name = ""
    nip05 = ""
    lud16 = ""
    try:
        pk = PublicKey.parse(npub)
    except:
        return "", "", ""
    opts = (Options().wait_for_send(False).send_timeout(datetime.timedelta(seconds=5)))
    keys = Keys.parse(check_and_set_private_key("test_client"))
    signer = NostrSigner.keys(keys)
    client = ClientBuilder().signer(signer).opts(opts).build()
    await client.add_relay("wss://relay.damus.io")
    await client.add_relay("wss://relay.primal.net")
    await client.add_relay("wss://purplepag.es")
    await client.connect()

    profile_filter = Filter().kind(Kind(0)).author(pk).limit(1)

    events = await client.get_events_of([profile_filter], relay_timeout)
    if len(events) > 0:
        try:
            profile = json.loads(events[0].content())
            if profile.get("name"):
                name = profile['name']
            if profile.get("nip05"):
                nip05 = profile['nip05']
            if profile.get("lud16"):
                lud16 = profile['lud16']
        except Exception as e:
            print(e)
    await client.shutdown()
    return name, nip05, lud16


async def print_results(graph, index_map, show_results_num, getmetadata=True):
    for item in islice(graph, show_results_num):
        key = next((PublicKey.parse(pubkey).to_bech32() for pubkey, id in index_map.items() if id == item), None)
        name = ""
        if getmetadata:
            name, nip05, lud16 = await get_metadata(key)
        print(name + "(" + key + ") " + str(graph[item]))


async def convert_index_to_hex(graph, index_map, show_results_num):
    result = {}
    for item in islice(graph, show_results_num):
        key = next((pubkey for pubkey, id in index_map.items() if id == item), None)
        result[key] = graph[item]

    return result


def test():
    # WARNING, DEPENDING ON DEPTH THIS TAKES LONG
    user = '3bf0c63fcb93463407af97a5e5ee64fa883d107ef9e558472c4eb9aaaefa459d'
    index_map, network_graph = asyncio.run(build_wot_network(user, depth=2, max_batch=500, max_time_request=10))
    save_network(index_map, network_graph, user)



# Copyright © 2023 Apple Inc.

import mlx.core as mx

from .config import DiffusionConfig


def _linspace(a, b, num):
    x = mx.arange(0, num) / (num - 1)
    return (b - a) * x + a


def _interp(y, x_new):
    """Interpolate the function defined by (arange(0, len(y)), y) at positions x_new."""
    x_low = x_new.astype(mx.int32)
    x_high = mx.minimum(x_low + 1, len(y) - 1)

    y_low = y[x_low]
    y_high = y[x_high]
    delta_x = x_new - x_low
    y_new = y_low * (1 - delta_x) + delta_x * y_high

    return y_new


class SimpleEulerSampler:
    """A simple Euler integrator that can be used to sample from our diffusion models.

    The method ``step()`` performs one Euler step from x_t to x_t_prev.
    """

    def __init__(self, config: DiffusionConfig):
        # Compute the noise schedule
        if config.beta_schedule == "linear":
            betas = _linspace(
                config.beta_start, config.beta_end, config.num_train_steps
            )
        elif config.beta_schedule == "scaled_linear":
            betas = _linspace(
                config.beta_start ** 0.5, config.beta_end ** 0.5, config.num_train_steps
            ).square()
        else:
            raise NotImplementedError(f"{config.beta_schedule} is not implemented.")

        alphas = 1 - betas
        alphas_cumprod = mx.cumprod(alphas)

        self._sigmas = mx.concatenate(
            [mx.zeros(1), ((1 - alphas_cumprod) / alphas_cumprod).sqrt()]
        )

    def sample_prior(self, shape, dtype=mx.float32, key=None):
        noise = mx.random.normal(shape, key=key)
        return (
                noise * self._sigmas[-1] * (self._sigmas[-1].square() + 1).rsqrt()
        ).astype(dtype)

    def sigmas(self, t):
        return _interp(self._sigmas, t)

    def timesteps(self, num_steps: int, dtype=mx.float32):
        steps = _linspace(len(self._sigmas) - 1, 0, num_steps + 1).astype(dtype)
        return list(zip(steps, steps[1:]))

    def step(self, eps_pred, x_t, t, t_prev):
        sigma = self.sigmas(t).astype(eps_pred.dtype)
        sigma_prev = self.sigmas(t_prev).astype(eps_pred.dtype)

        dt = sigma_prev - sigma
        x_t_prev = (sigma.square() + 1).sqrt() * x_t + eps_pred * dt

        x_t_prev = x_t_prev * (sigma_prev.square() + 1).rsqrt()

        return x_t_prev



AVOID_OUTBOX_RELAY_LIST = ["wss://nostrelay.yeghro.site", "wss://nostr.wine", "wss://filter.nostr.wine",
                           "wss://nostr21.com", "wss://nostr.bitcoiner.social", "wss://nostr.orangepill.dev",
                           "wss://brb.io",
                           "wss://relay.lnpay.me", "wss://relay.snort.social", "wss://relay.minds.com/nostr/v1/ws",
                           "ws://elitedesk:4848",
                           "wss://nostr-pub.semisol.dev", "wss://mostr.mostr.pub", "wss://relay.mostr.pub",
                           "wss://minds.com",
                           "wss://yabu.me", "wss://relay.yozora.world", "wss://filter.nostr.wine/?global=all",
                           "wss://eden.nostr.land",
                           "wss://relay.orangepill.ovh", "wss://nostr.jcloud.es", "wss://af.purplerelay.com",
                           "wss://za.purplerelay.com",
                           "wss://relay.nostrich.land", "wss://relay.nostrplebs.com", "wss://relay.nostrich.land",
                           "ws://elitedesk.local:4848",
                           "wss://rss.nos.social", "wss://atlas.nostr.land", "wss://puravida.nostr.land",
                           "wss://nostr.inosta.cc",
                           "wss://relay.orangepill.dev", "wss://no.str.cr", "wss://nostr.milou.lol",
                           "wss://relay.nostr.com.au",
                           "wss://puravida.nostr.land", "wss://atlas.nostr.land", "wss://nostr-pub.wellorder.net",
                           "wss://relay.current.fyi",
                           "wss://nostr.thesamecat.io", "wss://nostr.plebchain.org", "wss://relay.noswhere.com",
                           "wss://nostr.uselessshit.co",
                           "wss://bitcoiner.social", "wss://relay.stoner.com", "wss://nostr.l00p.org",
                           "wss://relay.nostr.ro", "wss://nostr.kollider.xyz",
                           "wss://relay.valera.co", "wss://relay.austrich.net", "wss://relay.nostrich.de",
                           "wss://nostr.azte.co", "wss://nostr-relay.schnitzel.world",
                           "wss://relay.nostriches.org", "wss://happytavern.co", "wss://onlynotes.lol",
                           "wss://offchain.pub", "wss://purplepag.es", "wss://relay.plebstr.com",
                           "wss://poster.place/relay", "wss://relayable.org", "wss://bbb.santos.lol",
                           "wss://relay.bitheaven.social", "wss://theforest.nostr1.com", "wss://at.nostrworks.com",
                           "wss://relay.nostrati.com", "wss://purplerelay.com", "wss://hist.nostr.land",
                           "wss://creatr.nostr.wine", "ws://localhost:4869", "wss://pleb.cloud",
                           "wss://pyramid.fiatjaf.com", "wss://relay.nos.social", "wss://nostr.thank.eu",
                           "wss://inbox.nostr.wine", "wss://relay.pleb.to", "wss://welcome.nostr.wine",
                           "wss://relay.nostrview.com", "wss://nostr.land", "wss://eu.purplerelay.com",
                           "wss://xmr.usenostr.org", "wss://nostr-relay.app", "ws://umbrel:4848", "wss://umbrel:4848",
                           "wss://fiatjaf.com", "wss://nostr-relay.wlvs.space", "wss://relayer.fiatjaf.com",
                           "wss://nostr.yuv.al", "wss://relay.nostr.band", "wss://nostr.massmux.com",
                           "wss://nostr-01.bolt.observer", "wss://nostr1.tunnelsats.com", "wss://relay.nostr.ch",
                           "wss://relay.nostr.io", "wss://nostr.thank.eu", "wss://nostr.bitcoinplebs.de",
                           "wss://adult.18plus.social", "wss://bostr.online", "wss://relay.current.fyi",
                           "wss://nosdrive.app/relay", "wss://studio314.nostr1.com", "wss://relay.nostrbr.online"
                           ]



import asyncio
import json
from datetime import timedelta

from nostr_sdk import PublicKey, Options, Keys, Client, NostrSigner

from nostr_dvm.interfaces.dvmtaskinterface import DVMTaskInterface
from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nip89_utils import NIP89Config, nip89_fetch_events_pubkey
from nostr_dvm.utils.output_utils import PostProcessFunctionType


async def build_client(config):
    opts = (Options().wait_for_send(True).send_timeout(timedelta(seconds=config.RELAY_TIMEOUT))
            .skip_disconnected_relays(True))
    keys = Keys.parse(config.PRIVATE_KEY)
    signer = NostrSigner.keys(keys)
    client = Client.with_opts(signer, opts)

    for relay in config.RELAY_LIST:
        await client.add_relay(relay)
    await client.connect()
    return client


def build_external_dvm(pubkey, task, kind, fix_cost, per_unit_cost, config,
                       external_post_process=PostProcessFunctionType.NONE):
    dvm_config = DVMConfig()
    dvm_config.PUBLIC_KEY = PublicKey.from_hex(pubkey).to_hex()
    dvm_config.FIX_COST = fix_cost
    dvm_config.PER_UNIT_COST = per_unit_cost
    dvm_config.EXTERNAL_POST_PROCESS_TYPE = external_post_process

    client = asyncio.run(build_client(config))

    nip89content_str = asyncio.run(nip89_fetch_events_pubkey(client, pubkey, kind))
    name = "External DVM"
    image = "https://image.nostr.build/c33ca6fc4cc038ca4adb46fdfdfda34951656f87ee364ef59095bae1495ce669.jpg"
    about = "An External DVM with no info"
    nip90params = {}
    encryption_supported = False
    cashu_accepted = False

    if nip89content_str is not None:
        print(nip89content_str)
        nip89content = json.loads(nip89content_str)
        if nip89content.get("name"):
            name = nip89content.get("name")
        if nip89content.get("image"):
            image = nip89content.get("image")
        if nip89content.get("about"):
            about = nip89content.get("about")
        if nip89content.get("nip90Params"):
            nip90params = nip89content["nip90Params"]
        if nip89content.get("encryptionSupported"):
            encryption_supported = nip89content["encryptionSupported"]
        if nip89content.get("cashuAccepted"):
            cashu_accepted = nip89content["cashuAccepted"]
    else:
        print("No NIP89 set for " + name)
    nip89info = {
        "name": name,
        "image": image,
        "about": about,
        "encryptionSupported": encryption_supported,
        "cashuAccepted": cashu_accepted,
        "nip90Params": nip90params
    }
    nip89config = NIP89Config()
    nip89config.KIND = kind
    nip89config.CONTENT = json.dumps(nip89info)

    interface = DVMTaskInterface(name=name, dvm_config=dvm_config, nip89config=nip89config, task=task)
    interface.SUPPORTS_ENCRYPTION = encryption_supported
    interface.ACCEPTS_CASHU = cashu_accepted

    return interface



import asyncio
import json
from datetime import timedelta
from pathlib import Path

import dotenv
from nostr_sdk import Keys, Client, Tag, EventBuilder, Filter, HandleNotification, Timestamp, nip04_decrypt, \
    nip04_encrypt, EventId, Options, PublicKey, Event, NostrSigner, Nip19Event, SecretKey, Kind

from nostr_dvm.utils import definitions, dvmconfig
from nostr_dvm.utils.definitions import relay_timeout
from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.gallery_utils import gallery_announce_list
from nostr_dvm.utils.nip89_utils import NIP89Config
from nostr_dvm.utils.nostr_utils import check_and_set_private_key


async def test():
    relay_list = dvmconfig.DVMConfig.RELAY_LIST
    keys = Keys.parse(check_and_set_private_key("test_client"))
    wait_for_send = False
    skip_disconnected_relays = True
    opts = (Options().wait_for_send(wait_for_send).send_timeout(timedelta(seconds=5))
            .skip_disconnected_relays(skip_disconnected_relays))

    signer = NostrSigner.keys(keys)
    client = Client.with_opts(signer, opts)

    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()

    await test_referred_events(client, "c70fbd4dbaad22c427d4359981d3bdddd3971ed1a38227ca2f8e5e760f58103c",
                               definitions.EventDefinitions.ANY_RESULT)

    # shows kind 7000 reaction but not kind 6300 result (d05e7ae9271fe2d8968cccb67c01e3458dbafa4a415e306d49b22729b088c8a1)
    await test_referred_events(client, "5635e5dd930b3c831f6ab1e348bb488f3c9aca2f13190e93ab5e5e1e1ba1835e",
                               definitions.EventDefinitions.ANY_RESULT)

    bech32evnt = EventId.from_hex("5635e5dd930b3c831f6ab1e348bb488f3c9aca2f13190e93ab5e5e1e1ba1835e").to_bech32()
    print(bech32evnt)

    test = Nip19Event.from_bech32(
        "nevent1qqsrjcpejsrlt3u7dy42y6rc97svrq9ver08xy4jr2ll55ynq3sxafcppamhxue69uhkummnw3ezumt0d5pzpmnqx2pla0zvxxcfjqeeysy29ll3mtmf4s3yff0y45r7egau080vqvzqqqqqqyu4q839")
    print(test.event_id().to_hex())

    nostruri = EventId.from_hex("5635e5dd930b3c831f6ab1e348bb488f3c9aca2f13190e93ab5e5e1e1ba1835e").to_nostr_uri()
    print(nostruri)

    await test_search_by_user_since_days(client,
                                         PublicKey.from_bech32(
                                             "npub1nxa4tywfz9nqp7z9zp7nr7d4nchhclsf58lcqt5y782rmf2hefjquaa6q8"), 60,
                                         "Bitcoin")


async def test_referred_events(client, event_id, kinds=None):
    if kinds is None:
        kinds = []

    if len(kinds) > 0:
        job_id_filter = Filter().kinds(kinds).event(EventId.from_hex(event_id))
    else:
        job_id_filter = Filter().event(EventId.from_hex(event_id))

    events = await client.get_events_of([job_id_filter], relay_timeout)

    if len(events) > 0:
        for event in events:
            print(event.as_json())
        return events[0]
    else:
        print("None")
        return None


async def test_gallery():
    relay_list = dvmconfig.DVMConfig.RELAY_LIST
    keys = Keys.parse(check_and_set_private_key("test_client"))
    wait_for_send = False
    skip_disconnected_relays = True
    opts = (Options().wait_for_send(wait_for_send).send_timeout(timedelta(seconds=5))
            .skip_disconnected_relays(skip_disconnected_relays))

    signer = NostrSigner.keys(keys)
    client = Client.with_opts(signer, opts)

    for relay in relay_list:
        await client.add_relay(relay)
    await client.connect()
    dvm_config = DVMConfig()
    dvm_config.NIP89 = NIP89Config()
    keys = Keys.parse(check_and_set_private_key("RTEST_ACCOUNT_PK"))
    print(keys.public_key().to_hex())
    dvm_config.NIP89.PK = keys.secret_key().to_hex()
    tagname = "url"
    tags = [

        Tag.parse([tagname, "https://i.nostr.build/xEZqV.jpg",
                   "3b0ec270394dc496f9f9c7db5c68a5b7f7311ff9080a51f1e8cb5f5cffc2c0b2", "wss://nostr.mom"]),
        # Tag.parse([tagname, "https://i.nostr.build/2RnXd.jpg", "dd6e5c2891fbe9f53bcaa351b48faeeedccd16e9541b508adcb2c16d11bceaaf", "wss://nostr.mom"]),
        # Tag.parse([tagname, "https://i.nostr.build/WG2Ra.jpg", "b2868e1ef93523ecf15b26e1cfdb6f252fe5074867d9c042fd6fcfbf07959193", "wss://nostr.mom"]),
        # Tag.parse([tagname, "https://i.nostr.build/M5keE.jpg", "489402bf3ec070e7ebf2ba459508d2e1a408c0adad02954470602f232026a37d", "wss://nostr.mom"]),
        Tag.parse([tagname, "https://v.nostr.build/M5kZ5.mp4",
                   "0e37cb0373189e01be3c744c0434e0c8559953910e44b05ed270313c47abe142", "wss://nostr.mom"]),
        Tag.parse([tagname, "https://i.nostr.build/vGLg7.jpg",
                   "102d1f411a9a2b4de37ef62cdd4943673b4941080a51a8fa8829cd9f1de46d13", "wss://nostr.mom"]),
        Tag.parse([tagname, "https://i.nostr.build/O4WxA.jpg",
                   "4022d4e893c224186bbef4414340e35cbf251c681bc84ab05446fec1d2ec67df", "wss://nostr.mom"]),
        Tag.parse([tagname, "https://i.nostr.build/M5a96.jpg",
                   "6f04dc6a2a05f710b9c6c6d09a02c5fe0174da9c95399d3d01963a784d195803", "wss://nostr.mom"]),
        Tag.parse([tagname, "https://i.nostr.build/Dj2Q4.jpg",
                   "737a169c245ce7957a8b6acf190c57d70256cc52630862f5ba0fd7315ef83425", "wss://nostr.mom"]),
        Tag.parse([tagname, "https://i.nostr.build/Dj2Q4.jpg",
                   "c3a3a8759502cb3c06d592e5715cad0826982a2ff60a0ae525e3f253ab9e462a", "wss://nostr.mom"]),
        # Tag.parse([tagname, "https://i.nostr.build/7G2G2.jpg", "015e71ded102e96d2b30f63dec0c04546d52a51f709709391af68d73f7502feb", "wss://nostr.mom"]),
        # Tag.parse([tagname, "https://i.nostr.build/XVLkd.jpg", "43da37c84113d4c0bdc60ae1c82cef9761ff7a2a1ef29b1cc26abfd4932786c5", "wss://nostr.mom"]),
        Tag.parse(["alt", "Profile Gallery List"])
    ]



    # await gallery_announce_list(tags, dvm_config, client)

    #evt =  EventBuilder.delete([EventId.parse("721ac7c7d9309b6d3e6728a7274f5a1f10096b4ab17068233bcfa05cb233e84a")],
    #                          reason="deleted").to_event(keys)
    #await client.send_event(evt)

    #key1 = Keys.parse("3e99f38a8e8c59ff3683cdc0942e26471c1aae9b225eb34dd410cb9d6dde93a6")
    #key2 = Keys.parse("c35006d2c1e5618bfc107030047b85da4f11638b2901b24193fa45160890d851")

    #encrypted_text = nip04_encrypt(key1.secret_key(), key2.public_key(), "Hello")
    #print(encrypted_text)
    #encrypted_text = "T+f58RV7ENaSLWZPgUEiWA==?iv=Ize4cInMI3xuWnL5I0nYFw=="

    #rsult = nip04_decrypt(key2.secret_key(), key1.public_key(), encrypted_text)
    #print(rsult)


async def test_search_by_user_since_days(client, pubkey, days, prompt):
    since_seconds = int(days) * 24 * 60 * 60
    dif = Timestamp.now().as_secs() - since_seconds
    since = Timestamp.from_secs(dif)

    filterts = Filter().search(prompt).author(pubkey).kinds([Kind(1)]).since(since)
    events = await client.get_events_of([filterts], relay_timeout)

    if len(events) > 0:
        for event in events:
            print(event.as_json())
        return events[0]
    else:
        print("None")
        return None


if __name__ == '__main__':
    env_path = Path('.env')
    if env_path.is_file():
        print(f'loading environment from {env_path.resolve()}')
        dotenv.load_dotenv(env_path, verbose=True, override=True)
    else:
        raise FileNotFoundError(f'.env file not found at {env_path} ')

    asyncio.run(test_gallery())




import asyncio

from nostr_sdk import Tag, Keys, EventBuilder, Kind, NostrSigner, Client

from nostr_dvm.utils.dvmconfig import DVMConfig
from nostr_dvm.utils.nostr_utils import send_event, check_and_set_private_key
from nostr_dvm.utils.print_utils import bcolors


async def create_reaction(keys, title, dtag):
    d_tag = Tag.parse(["d", dtag])
    title_tag = Tag.parse(["title", title])

    emoji_tags = []

    # add more if you want
    name = "ThugAmy"
    url = "https://image.nostr.build/ccc229cbe11f5a13a1cc7fd24e13ac53fc78f287ecce0d9a674807e2e20f6fd5.png"
    emoji_tag1 = Tag.parse(["emoji", name, url])
    emoji_tags.append(emoji_tag1)

    keys = Keys.parse(keys)
    content = ""
    event = EventBuilder(Kind(30030), content, [d_tag, title_tag] + emoji_tags).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    # We add the relays we defined above and told our DVM we would want to receive events to.
    for relay in DVMConfig().RELAY_LIST:
        await client.add_relay(relay)
    # We connect the client
    await client.connect()

    eventid = await send_event(event, client=client, dvm_config=DVMConfig())

    print(
        bcolors.BLUE + "[" + "Reaction" + "] Announced (" + eventid.id.to_nostr_uri() +
        " Hex: " + eventid.id.to_hex() + ")" + bcolors.ENDC)


async def delete_reaction(keys, eid: str, dtag: str):
    keys = Keys.parse(keys)
    e_tag = Tag.parse(["e", eid])
    a_tag = Tag.parse(
        ["a", "30030:" + keys.public_key().to_hex() + ":" + dtag])
    event = EventBuilder(Kind(5), "", [e_tag, a_tag]).to_event(keys)

    signer = NostrSigner.keys(keys)
    client = Client(signer)
    # We add the relays we defined above and told our DVM we would want to receive events to.
    for relay in DVMConfig().RELAY_LIST:
        await client.add_relay(relay)
    # We connect the client
    await client.connect()

    eventid = await send_event(event, client, DVMConfig())
    print(
        bcolors.BLUE + "[" + "Reaction" + "] deleted (" + eventid.id.to_nostr_uri() +
        " Hex: " + eventid.id.to_hex() + ")" + bcolors.ENDC)


keys = check_and_set_private_key("test_client")
eventid = "da05cefc512ad43363f84131343f5d2a80303ea3b9368b9ad7f010e07db37d90"

asyncio.run(create_reaction(keys=keys, title="ThugAmy", dtag="ThugAmy"))
# asyncio.run(delete_reaction(keys=keys, eid=eventid,  dtag="ThugAmy"))



import base64
import hashlib

from nostr_sdk import EventBuilder, Tag, Kind, Keys


def sha256sum(filename):
    with open(filename, 'rb', buffering=0) as f:
        return hashlib.file_digest(f, 'sha256').hexdigest()


async def generate_nip98_header(pkeys_hex, url="", kind="POST", filepath=""):
    keys = Keys.parse(pkeys_hex)
    utag = Tag.parse(["u", url])
    methodtag = Tag.parse(["method", kind])
    tags = [utag, methodtag]
    if kind == "POST":
        payloadtag = Tag.parse(["payload", sha256sum(filepath)])
        tags.append(payloadtag)
    event = EventBuilder(Kind(27235), "", tags).to_event(keys)

    encoded_nip98_event = base64.b64encode(event.as_json().encode('utf-8')).decode('utf-8')

    return "Nostr " + encoded_nip98_event



import base64
import json

import requests

from nostr_dvm.utils.database_utils import get_or_add_user
from nostr_dvm.utils.zap_utils import create_bolt11_ln_bits, create_bolt11_lud16

BASE_URL = "https://mint.minibits.cash/Bitcoin"


def parse_cashu(cashu_token: str):
    try:
        prefix = "cashuA"
        assert cashu_token.startswith(prefix), Exception(
            f"Token prefix not valid. Expected {prefix}."
        )
        if not cashu_token.endswith("="):
            cashu_token = str(cashu_token) + "=="
        print(cashu_token)
        token_base64 = cashu_token[len(prefix):].encode("utf-8")
        cashu = json.loads(base64.urlsafe_b64decode(token_base64))
        token = cashu["token"][0]
        proofs = token["proofs"]
        mint = token["mint"]
        total_amount = 0
        for proof in proofs:
            total_amount += proof["amount"]

        return proofs, mint, total_amount, None

    except Exception as e:
        print(e)
        return None, None, None, "Cashu Parser: " + str(e)


async def redeem_cashu(cashu, config, client, required_amount=0, update_self=False) -> (bool, str, int, int):
    proofs, mint, total_amount, message = parse_cashu(cashu)
    if message is not None:
        return False, message, 0, 0

    estimated_fees = max(int(total_amount * 0.02), 3)
    estimated_redeem_invoice_amount = total_amount - estimated_fees

    # Not sure if this the best way to go, we first create an invoice that we send to the mint, we catch the fees
    # for that invoice, and create another invoice with the amount without fees to melt.
    if config.LNBITS_INVOICE_KEY != "":
        invoice, paymenthash = create_bolt11_ln_bits(estimated_redeem_invoice_amount, config)
    else:

        user = await get_or_add_user(db=config.DB, npub=config.PUBLIC_KEY,
                                     client=client, config=config, update=update_self)
        invoice = create_bolt11_lud16(user.lud16, estimated_redeem_invoice_amount)
    print(invoice)
    if invoice is None:
        return False, "couldn't create invoice", 0, 0

    url = mint + "/checkfees"  # Melt cashu tokens at Mint
    json_object = {"pr": invoice}
    headers = {"Content-Type": "application/json; charset=utf-8"}
    request_body = json.dumps(json_object).encode('utf-8')
    request = requests.post(url, data=request_body, headers=headers)
    tree = json.loads(request.text)
    fees = tree["fee"]
    print("Fees on this mint are " + str(fees) + " Sats")
    redeem_invoice_amount = total_amount - fees
    if redeem_invoice_amount < required_amount:
        err = ("Token value (Payment: " + str(total_amount) + " Sats. Fees: " +
               str(fees) + " Sats) below required amount of  " + str(required_amount)
               + " Sats. Cashu token has not been claimed.")
        print("[" + config.NIP89.NAME + "] " + err)
        return False, err, 0, 0

    if config.LNBITS_INVOICE_KEY != "":
        invoice, paymenthash = create_bolt11_ln_bits(redeem_invoice_amount, config)
    else:

        user = await get_or_add_user(db=config.DB, npub=config.PUBLIC_KEY,
                                     client=client, config=config, update=update_self)
        invoice = create_bolt11_lud16(user.lud16, redeem_invoice_amount)
    print(invoice)

    try:
        url = mint + "/melt"  # Melt cashu tokens at Mint
        json_object = {"proofs": proofs, "pr": invoice}
        headers = {"Content-Type": "application/json; charset=utf-8"}
        request_body = json.dumps(json_object).encode('utf-8')
        request = requests.post(url, data=request_body, headers=headers)
        tree = json.loads(request.text)
        print(request.text)
        is_paid = tree["paid"] if tree.get("paid") else False
        print(is_paid)
        if is_paid:
            print("cashu token redeemed")
            return True, "success", redeem_invoice_amount, fees
        else:
            msg = tree.get("detail").split('.')[0].strip() if tree.get("detail") else None
            print(msg)
            return False, msg, redeem_invoice_amount, fees
    except Exception as e:
        print(e)

    return False, "", redeem_invoice_amount, fees



import datetime as datetime
import json
import os
import random
from types import NoneType

import emoji
import pandas
import requests
from nostr_sdk import Tag, PublicKey, EventId, Keys, EventBuilder, LogLevel
from pyupload.uploader import CatboxUploader

from nostr_dvm.utils.definitions import EventDefinitions
from nostr_dvm.utils.nip98_utils import generate_nip98_header
from nostr_dvm.utils.nostr_utils import send_event_outbox
from nostr_dvm.utils.print_utils import bcolors

'''
Post process results to either given output format or a Nostr readable plain text.
'''


class PostProcessFunctionType:
    NONE = 0
    LIST_TO_USERS = 1
    LIST_TO_EVENTS = 2


def post_process_result(anno, original_event):
    print("Post-processing...")
    if isinstance(anno, pandas.DataFrame):  # if input is an anno we parse it to required output format
        print("Pandas Dataframe...")
        has_output_tag = False
        output_format = "text/plain"

        for tag in original_event.tags():
            if tag.as_vec()[0] == "output":
                output_format = tag.as_vec()[1]
                has_output_tag = True
                print("requested output is " + str(output_format) + "...")

        if has_output_tag:
            print("Output Tag found: " + output_format)
            try:
                if output_format == "text/plain":
                    result = pandas_to_plaintext(anno)
                    result = replace_broken_words(
                        str(result).replace("\"", "").replace('[', "").replace(']',
                                                                               "").lstrip(None))
                    return result

                elif output_format == "text/vtt":
                    print(str(anno))
                    result = "WEBVTT\n\n"
                    for element in anno:
                        name = element["name"]  # name
                        start = float(element["from"])
                        convertstart = str(datetime.timedelta(seconds=start))
                        end = float(element["to"])
                        convertend = str(datetime.timedelta(seconds=end))
                        print(str(convertstart) + " --> " + str(convertend))
                        cleared_name = str(name).lstrip("\'").rstrip("\'")
                        result = result + str(convertstart) + " --> " + str(
                            convertend) + "\n" + cleared_name + "\n\n"
                    result = replace_broken_words(
                        str(result).replace("\"", "").replace('[', "").replace(']',
                                                                               "").lstrip(None))
                    return result

                elif output_format == "text/json" or output_format == "json":
                    # result = json.dumps(json.loads(anno.data.to_json(orient="records")))
                    result = replace_broken_words(json.dumps(anno.data.tolist()))
                    return result
                # TODO add more
                else:
                    print("Pandas Dataframe but output tag not supported.. falling back to default..")
                    result = pandas_to_plaintext(anno)
                    print(result)
                    result = str(result).replace("\"", "").replace('[', "").replace(']',
                                                                                    "").lstrip(None)
                    return result

            except Exception as e:
                print(e)
                result = replace_broken_words(str(anno.data))
                return result

        else:
            print("Pandas Dataframe but no output tag set.. falling back to default..")
            result = pandas_to_plaintext(anno)
            print(result)
            result = str(result).replace("\"", "").replace('[', "").replace(']',
                                                                            "").lstrip(None)
            return result
    elif isinstance(anno, NoneType):
        return "An error occurred"
    else:
        result = replace_broken_words(anno)  # TODO
        return result


def post_process_list_to_events(result):
    result_list = json.loads(result)
    result_str = ""
    if len(result_list) == 0:
        return "No results found"
    for tag in result_list:
        e_tag = Tag.parse(tag)
        result_str = result_str + "nostr:" + EventId.from_hex(
            e_tag.as_vec()[1]).to_bech32() + "\n"
    return result_str


def post_process_list_to_users(result):
    result_list = json.loads(result)
    result_str = ""
    if len(result_list) == 0:
        return "No results found"
    for tag in result_list:
        p_tag = Tag.parse(tag)
        result_str = result_str + "nostr:" + PublicKey.parse(
            p_tag.as_vec()[1]).to_bech32() + "\n"
    return result_str


def pandas_to_plaintext(anno):
    result = ""
    for each_row in anno['name']:
        if each_row is not None:
            for i in str(each_row).split('\n'):
                result = result + i + "\n"
    return result


'''
Convenience function to replace words like Noster with Nostr
'''


def replace_broken_words(text):
    result = (text.replace("Noster", "Nostr").replace("Nostra", "Nostr").replace("no stir", "Nostr").
              replace("Nostro", "Nostr").replace("Impub", "npub").replace("sets", "Sats"))
    return result


'''
Function to upload to Nostr.build and if it fails to Nostrfiles.dev
Larger files than these hosters allow and fallback is catbox currently.
Will probably need to switch to another system in the future.
'''


async def upload_media_to_hoster(filepath: str, key_hex=None, fallback=True):
    if key_hex is None:
        # If we don't pass a key we use the key from .env
        if os.getenv("NOSTR_BUILD_ACCOUNT_PK"):
            key_hex = Keys.parse(os.getenv("NOSTR_BUILD_ACCOUNT_PK")).secret_key().to_hex()
        else:
            print("No Hex key set, using catbox")
            uploader = CatboxUploader(filepath)
            result = uploader.execute()
            return result

    print("Uploading Media: " + filepath)
    try:
        files = {'file': open(filepath, 'rb')}
        file_stats = os.stat(filepath)
        sizeinmb = file_stats.st_size / (1024 * 1024)
        print("Filesize of Uploaded media: " + str(sizeinmb) + " Mb.")

        limitinmb = await request_nostr_build_limit(key_hex)

        if sizeinmb > limitinmb:
            if fallback:
                print("Filesize over Nostr.build limit, using paid account")
                key_hex = Keys.parse(os.getenv("NOSTR_BUILD_ACCOUNT_PK")).secret_key().to_hex()
                limitinmb = await request_nostr_build_limit(key_hex)
                if sizeinmb > limitinmb:
                    return await upload_nostr_build(key_hex, files, filepath)
                else:
                    print("Filesize over paid Nostr.build limit, using catbox")
                    uploader = CatboxUploader(filepath)
                    result = uploader.execute()
                    return result

            else:
                print("Filesize over Nostr.build limit, using catbox")
                uploader = CatboxUploader(filepath)
                result = uploader.execute()
                return result
        else:
            return await upload_nostr_build(key_hex, files, filepath)

    except Exception as e:
        print(e)
        try:
            # on error we fallback to catbox nevertheless
            uploader = CatboxUploader(filepath)
            result = uploader.execute()
            print(result)
            return result
        except:
            raise Exception("Upload not possible, all hosters didn't work or couldn't generate output")


async def upload_nostr_build(pkey, files, filepath):
    url = 'https://nostr.build/api/v2/upload/files'
    auth = await generate_nip98_header(pkey, url, "POST", filepath)
    headers = {'authorization': auth}

    response = requests.post(url, files=files, headers=headers)
    print(response.text)
    json_object = json.loads(response.text)
    result = json_object["data"][0]["url"]
    return result


async def request_nostr_build_limit(pkey):
    url = 'https://nostr.build/api/v2/upload/limit'
    auth = await generate_nip98_header(pkey, url, "GET")
    headers = {'authorization': auth}
    response = requests.get(url, headers=headers)
    json_object = json.loads(response.text)
    limit = float(json_object["data"]["limit"])
    limitinmb = limit / (1024 * 1024)
    print("Limit: " + str(limitinmb) + " MB")
    return limitinmb


def build_status_reaction(status, task, amount, content, dvm_config):
    alt_description = "This is a reaction to a NIP90 DVM AI task. "

    if status == "processing":
        if content is not None:
            if isinstance(content, list) or isinstance(content, dict):
                message = random.choice(content)
            else:
                message = content
            alt_description = message
            reaction = alt_description

        else:
            alt_description = "NIP90 DVM task " + task + " started processing. "
            reaction = alt_description + emoji.emojize(":thumbs_up:")
    elif status == "success":
        alt_description = "NIP90 DVM task " + task + " finished successfully. "
        reaction = alt_description + emoji.emojize(":call_me_hand:")
    elif status == "chain-scheduled":
        alt_description = "NIP90 DVM task " + task + " Chain Task scheduled"
        reaction = alt_description + emoji.emojize(":thumbs_up:")
    elif status == "error":
        alt_description = "NIP90 DVM task " + task + " had an error. "
        if content is None:
            reaction = alt_description + emoji.emojize(":thumbs_down:")
        else:
            reaction = alt_description + emoji.emojize(":thumbs_down:") + " " + content

    elif status == "payment-required":
        alt_description = "NIP90 DVM task " + task + " requires payment of min " + str(
            amount) + " Sats. "
        reaction = alt_description + emoji.emojize(":orange_heart:")

    elif status == "subscription-required":
        if content is not None and content != "":
            alt_description = content
            reaction = alt_description

        else:
            alt_description = "NIP90 DVM task " + task + " requires payment for subscription"
            reaction = alt_description + emoji.emojize(":orange_heart:")

    elif status == "subscription-active":
        if content is not None and content != "":
            alt_description = content
            reaction = alt_description

        else:
            alt_description = "NIP90 DVM task " + task + " has active subscription"
            reaction = alt_description + emoji.emojize(":orange_heart:")



    elif status == "payment-rejected":
        alt_description = "NIP90 DVM task " + task + " payment is below required amount of " + str(
            amount) + " Sats. "
        reaction = alt_description + emoji.emojize(":thumbs_down:")
    elif status == "user-blocked-from-service":
        alt_description = "NIP90 DVM task " + task + " can't be performed. User has been blocked from Service. "
        reaction = alt_description + emoji.emojize(":thumbs_down:")
    else:
        reaction = emoji.emojize(":thumbs_down:")

    return alt_description, reaction


async def send_job_status_reaction(original_event_id_hex, original_event_author_hex, client, dvm_config,
                                   content=None,
                                   status="processing", user=None):
    alt_description, reaction = build_status_reaction(status, "generic", 0, content, dvm_config)

    e_tag = Tag.parse(["e", original_event_id_hex])
    p_tag = Tag.parse(["p", original_event_author_hex])
    alt_tag = Tag.parse(["alt", content])
    status_tag = Tag.parse(["status", status])

    reply_tags = [e_tag, alt_tag, status_tag]

    reply_tags.append(p_tag)
    content = reaction

    keys = Keys.parse(dvm_config.PRIVATE_KEY)
    reaction_event = EventBuilder(EventDefinitions.KIND_FEEDBACK, str(content), reply_tags).to_event(keys)
    # send_event(reaction_event, client=self.client, dvm_config=self.dvm_config)
    await send_event_outbox(reaction_event, client=client, dvm_config=dvm_config)

    if dvm_config.LOGLEVEL.value >= LogLevel.DEBUG.value:
        print(bcolors.YELLOW + "[" + dvm_config.NIP89.NAME + "]" + " Sent Kind " + str(
            EventDefinitions.KIND_FEEDBACK.as_u16()) + " Reaction: " + status + " " + reaction_event.as_json() + bcolors.ENDC)
    elif dvm_config.LOGLEVEL.value >= LogLevel.INFO.value:
        print(bcolors.YELLOW + "[" + dvm_config.NIP89.NAME + "]" + " Sent Kind " + str(
            EventDefinitions.KIND_FEEDBACK.as_u16()) + " Reaction: " + status + " " + reaction_event.id().to_hex() + bcolors.ENDC)

    return reaction_event.as_json()



# Copyright © 2023 Apple Inc.

import math
from typing import List

import mlx.core as mx
import mlx.nn as nn

from .config import AutoencoderConfig
from .unet import ResnetBlock2D, upsample_nearest


class Attention(nn.Module):
    """A single head unmasked attention for use with the VAE."""

    def __init__(self, dims: int, norm_groups: int = 32):
        super().__init__()

        self.group_norm = nn.GroupNorm(norm_groups, dims, pytorch_compatible=True)
        self.query_proj = nn.Linear(dims, dims)
        self.key_proj = nn.Linear(dims, dims)
        self.value_proj = nn.Linear(dims, dims)
        self.out_proj = nn.Linear(dims, dims)

    def __call__(self, x):
        B, H, W, C = x.shape

        y = self.group_norm(x)

        queries = self.query_proj(y).reshape(B, H * W, C)
        keys = self.key_proj(y).reshape(B, H * W, C)
        values = self.value_proj(y).reshape(B, H * W, C)

        scale = 1 / math.sqrt(queries.shape[-1])
        scores = (queries * scale) @ keys.transpose(0, 2, 1)
        attn = mx.softmax(scores, axis=-1)
        y = (attn @ values).reshape(B, H, W, C)

        y = self.out_proj(y)
        x = x + y

        return x


class EncoderDecoderBlock2D(nn.Module):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            num_layers: int = 1,
            resnet_groups: int = 32,
            add_downsample=True,
            add_upsample=True,
    ):
        super().__init__()

        # Add the resnet blocks
        self.resnets = [
            ResnetBlock2D(
                in_channels=in_channels if i == 0 else out_channels,
                out_channels=out_channels,
                groups=resnet_groups,
            )
            for i in range(num_layers)
        ]

        # Add an optional downsampling layer
        if add_downsample:
            self.downsample = nn.Conv2d(
                out_channels, out_channels, kernel_size=3, stride=2, padding=1
            )

        # or upsampling layer
        if add_upsample:
            self.upsample = nn.Conv2d(
                out_channels, out_channels, kernel_size=3, stride=1, padding=1
            )

    def __call__(self, x):
        for resnet in self.resnets:
            x = resnet(x)

        if "downsample" in self:
            x = self.downsample(x)

        if "upsample" in self:
            x = self.upsample(upsample_nearest(x))

        return x


class Encoder(nn.Module):
    """Implements the encoder side of the Autoencoder."""

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            block_out_channels: List[int] = [64],
            layers_per_block: int = 2,
            resnet_groups: int = 32,
    ):
        super().__init__()

        self.conv_in = nn.Conv2d(
            in_channels, block_out_channels[0], kernel_size=3, stride=1, padding=1
        )

        channels = [block_out_channels[0]] + list(block_out_channels)
        self.down_blocks = [
            EncoderDecoderBlock2D(
                in_channels,
                out_channels,
                num_layers=layers_per_block,
                resnet_groups=resnet_groups,
                add_downsample=i < len(block_out_channels) - 1,
                add_upsample=False,
            )
            for i, (in_channels, out_channels) in enumerate(zip(channels, channels[1:]))
        ]

        self.mid_blocks = [
            ResnetBlock2D(
                in_channels=block_out_channels[-1],
                out_channels=block_out_channels[-1],
                groups=resnet_groups,
            ),
            Attention(block_out_channels[-1], resnet_groups),
            ResnetBlock2D(
                in_channels=block_out_channels[-1],
                out_channels=block_out_channels[-1],
                groups=resnet_groups,
            ),
        ]

        self.conv_norm_out = nn.GroupNorm(
            resnet_groups, block_out_channels[-1], pytorch_compatible=True
        )
        self.conv_out = nn.Conv2d(block_out_channels[-1], out_channels, 3, padding=1)

    def __call__(self, x):
        x = self.conv_in(x)

        for l in self.down_blocks:
            x = l(x)

        x = self.mid_blocks[0](x)
        x = self.mid_blocks[1](x)
        x = self.mid_blocks[2](x)

        x = self.conv_norm_out(x)
        x = nn.silu(x)
        x = self.conv_out(x)

        return x


class Decoder(nn.Module):
    """Implements the decoder side of the Autoencoder."""

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            block_out_channels: List[int] = [64],
            layers_per_block: int = 2,
            resnet_groups: int = 32,
    ):
        super().__init__()

        self.conv_in = nn.Conv2d(
            in_channels, block_out_channels[-1], kernel_size=3, stride=1, padding=1
        )

        self.mid_blocks = [
            ResnetBlock2D(
                in_channels=block_out_channels[-1],
                out_channels=block_out_channels[-1],
                groups=resnet_groups,
            ),
            Attention(block_out_channels[-1], resnet_groups),
            ResnetBlock2D(
                in_channels=block_out_channels[-1],
                out_channels=block_out_channels[-1],
                groups=resnet_groups,
            ),
        ]

        channels = list(reversed(block_out_channels))
        channels = [channels[0]] + channels
        self.up_blocks = [
            EncoderDecoderBlock2D(
                in_channels,
                out_channels,
                num_layers=layers_per_block,
                resnet_groups=resnet_groups,
                add_downsample=False,
                add_upsample=i < len(block_out_channels) - 1,
            )
            for i, (in_channels, out_channels) in enumerate(zip(channels, channels[1:]))
        ]

        self.conv_norm_out = nn.GroupNorm(
            resnet_groups, block_out_channels[0], pytorch_compatible=True
        )
        self.conv_out = nn.Conv2d(block_out_channels[0], out_channels, 3, padding=1)

    def __call__(self, x):
        x = self.conv_in(x)

        x = self.mid_blocks[0](x)
        x = self.mid_blocks[1](x)
        x = self.mid_blocks[2](x)

        for l in self.up_blocks:
            x = l(x)

        x = self.conv_norm_out(x)
        x = nn.silu(x)
        x = self.conv_out(x)

        return x


class Autoencoder(nn.Module):
    """The autoencoder that allows us to perform diffusion in the latent space."""

    def __init__(self, config: AutoencoderConfig):
        super().__init__()

        self.latent_channels = config.latent_channels_in
        self.scaling_factor = config.scaling_factor
        self.encoder = Encoder(
            config.in_channels,
            config.latent_channels_out,
            config.block_out_channels,
            config.layers_per_block,
            resnet_groups=config.norm_num_groups,
        )
        self.decoder = Decoder(
            config.latent_channels_in,
            config.out_channels,
            config.block_out_channels,
            config.layers_per_block + 1,
            resnet_groups=config.norm_num_groups,
        )

        self.quant_proj = nn.Linear(
            config.latent_channels_out, config.latent_channels_out
        )
        self.post_quant_proj = nn.Linear(
            config.latent_channels_in, config.latent_channels_in
        )

    def decode(self, z):
        return self.decoder(self.post_quant_proj(z))

    def __call__(self, x, key=None):
        x = self.encoder(x)
        x = self.quant_proj(x)

        mean, logvar = x.split(2, axis=-1)
        std = mx.exp(0.5 * logvar)
        z = mx.random.normal(mean.shape, key=key) * std + mean

        x_hat = self.decode(z)

        return dict(x_hat=x_hat, z=z, mean=mean, logvar=logvar)
